import random
import time
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Sequence

import pandas as pd
import pytest
from confit import validate_arguments
from spacy.tokens import Doc

import edsnlp.data
import edsnlp.processing
from edsnlp.data.converters import get_current_tokenizer
from edsnlp.processing.multiprocessing import get_dispatch_schedule

try:
    import torch.nn
except ImportError:
    torch = None


docs = [
    {
        "note_id": 1234,
        "note_text": "This is a test.",
        "entities": [
            {
                "note_nlp_id": 0,
                "start_char": 0,
                "end_char": 4,
                "lexical_variant": "This",
                "note_nlp_source_value": "test",
                "negation": True,
            },
            {
                "note_nlp_id": 1,
                "start_char": 5,
                "end_char": 7,
                "lexical_variant": "is",
                "note_nlp_source_value": "test",
            },
        ],
    },
    {
        "note_id": 0,
        "note_text": "This is an empty document.",
        "entities": None,
    },
]


@pytest.mark.parametrize(
    "reader_format,reader_converter,backend,writer_format,writer_converter,worker_io",
    [
        ("pandas", "omop", "simple", "pandas", "omop", False),
        ("pandas", "omop", "multiprocessing", "pandas", "omop", False),
        ("pandas", "omop", "spark", "pandas", "omop", False),
        ("parquet", "omop", "simple", "parquet", "omop", False),
        ("parquet", "omop", "multiprocessing", "parquet", "omop", False),
        ("parquet", "omop", "spark", "parquet", "omop", False),
        ("parquet", "omop", "multiprocessing", "parquet", "omop", True),
        ("parquet", "omop", "spark", "parquet", "omop", True),
        ("parquet", "omop", "multiprocessing", "iterable", None, False),
    ],
)
def test_end_to_end(
    reader_format,
    reader_converter,
    backend,
    writer_format,
    writer_converter,
    worker_io,
    nlp_eds,
    tmp_path,
):
    nlp = nlp_eds
    rsrc = Path(__file__).parent.parent.resolve() / "resources"
    if reader_format == "pandas":
        pandas_dataframe = pd.DataFrame(docs)
        data = edsnlp.data.from_pandas(
            pandas_dataframe,
            converter=reader_converter,
        )
    elif reader_format == "parquet":
        data = edsnlp.data.read_parquet(
            rsrc / "docs.parquet",
            converter=reader_converter,
            read_in_worker=worker_io,
        )
    else:
        raise ValueError(reader_format)

    data = data.map_batches(lambda x: sorted(x, key=len), batch_size=2)
    data = data.map_pipeline(nlp)

    data = data.set_processing(
        backend=backend,
        show_progress=True,
        batch_by="words",
        batch_size=2,
    )

    if writer_format == "pandas":
        data.to_pandas(converter=writer_converter)
    elif writer_format == "parquet":
        data.write_parquet(
            tmp_path,
            converter=writer_converter,
            write_in_worker=worker_io,
        )
    elif writer_format == "iterable":
        list(data)
    else:
        raise ValueError(writer_format)


def test_multiprocessing_backend(frozen_ml_nlp):
    texts = ["Ceci est un exemple", "Ceci est un autre exemple"]
    docs = list(
        frozen_ml_nlp.pipe(
            texts * 20,
            batch_size=2,
        ).set_processing(
            backend="multiprocessing",
            num_cpu_workers=-1,
            sort_chunks=True,
            chunk_size=2,
            batch_by="words",
            show_progress=True,
        )
    )
    assert len(docs) == 40


def error_pipe(doc: Doc):
    if doc._.note_id == "text-3":
        raise ValueError("error")
    return doc


@pytest.mark.parametrize(
    "backend,deterministic",
    [
        ("simple", True),
        ("multiprocessing", True),
        ("multiprocessing", False),
        ("spark", True),
    ],
)
def test_multiprocessing_gpu_stub_backend(frozen_ml_nlp, backend, deterministic):
    text1 = "Ceci est un exemple"
    text2 = "Ceci est un autre exemple"
    stream = frozen_ml_nlp.pipe(
        chain.from_iterable(
            [
                text1,
                text2,
            ]
            for i in range(2)
        ),
    )
    if backend == "simple":
        pass
    elif backend == "multiprocessing":
        stream = stream.set_processing(
            batch_size=2,
            num_gpu_workers=1,
            num_cpu_workers=1,
            gpu_worker_devices=["cpu"],
            deterministic=deterministic,
        )
    elif backend == "spark":
        stream = stream.set_processing(backend="spark")
    list(stream)


def test_multiprocessing_gpu_stub_multi_cpu_deterministic_backend(frozen_ml_nlp):
    text1 = "Exemple"
    text2 = "Ceci est un autre exemple"
    text3 = "Ceci est un très long exemple ! Regardez tous ces mots !"
    texts = [text1, text2, text3] * 100
    random.Random(42).shuffle(texts)
    stream = frozen_ml_nlp.pipe(iter(texts))
    stream = stream.set_processing(
        batch_size="15 words",
        num_gpu_workers=1,
        num_cpu_workers=2,
        deterministic=True,
        # show_progress=True,
        # just to test in gpu-less environments
        gpu_worker_devices=["cpu"],
    )
    list(stream)


@pytest.mark.parametrize("wait", [True, False])
def test_multiprocessing_gpu_stub_wait(frozen_ml_nlp, wait):
    text1 = "Ceci est un exemple"
    text2 = "Ceci est un autre exemple"
    it = iter(
        frozen_ml_nlp.pipe(
            chain.from_iterable(
                [
                    text1,
                    text2,
                ]
                for i in range(2)
            ),
        ).set_processing(
            batch_size=2,
            num_gpu_workers=1,
            num_cpu_workers=1,
            gpu_worker_devices=["cpu"],
        )
    )
    if wait:
        time.sleep(5)
    list(it)


def simple_converter(obj):
    tok = get_current_tokenizer()
    doc = tok(obj["content"])
    doc._.note_id = obj["id"]
    return doc


def test_iterable_error():
    class Gen:
        def __iter__(self):
            for i in range(5):
                if i == 3:
                    raise ValueError("error")
                yield {"content": f"text-{i}", "id": f"text-{i}"}

    with pytest.raises(ValueError):
        list(
            edsnlp.data.from_iterable(Gen(), converter=simple_converter).set_processing(
                num_cpu_workers=2
            )
        )


def test_multiprocessing_rb_error(ml_nlp):
    text1 = "Ceci est un exemple"
    text2 = "Ceci est un autre exemple"
    ml_nlp.add_pipe(error_pipe, name="error", after="sentences")
    with pytest.raises(ValueError):
        docs = edsnlp.data.from_iterable(
            chain.from_iterable(
                [
                    {"content": text1, "id": f"text-{i}"},
                    {"content": text2, "id": f"other-text-{i}"},
                ]
                for i in range(5)
            ),
            converter=simple_converter,
        ).map(lambda x: time.sleep(0.2) or x)
        docs = ml_nlp.pipe(
            docs,
            n_process=2,
            batch_size=2,
        )
        list(docs)


if torch is not None:
    from edsnlp.core.torch_component import TorchComponent

    class DeepLearningError(TorchComponent):
        def __init__(self, *args, **kwargs):
            super().__init__()

        def preprocess(self, doc):
            return {"num_words": len(doc), "doc_id": doc._.note_id}

        def collate(self, batch):
            return {
                "num_words": torch.tensor(batch["num_words"]),
                "doc_id": batch["doc_id"],
            }

        def forward(self, batch):
            if "text-1" in batch["doc_id"]:
                raise RuntimeError("Deep learning error")
            return {}


@pytest.mark.skipif(torch is None, reason="torch not installed")
def test_multiprocessing_ml_error(ml_nlp):
    text1 = "Ceci est un exemple"
    text2 = "Ceci est un autre exemple"
    ml_nlp.add_pipe(
        DeepLearningError(pipeline=ml_nlp),
        name="error",
        after="sentences",
    )

    with pytest.raises(RuntimeError) as e:
        docs = edsnlp.data.from_iterable(
            chain.from_iterable(
                [
                    {"content": text1, "id": f"text-{i}"},
                    {"content": text2, "id": f"other-text-{i}"},
                ]
                for i in range(5)
            ),
            converter=simple_converter,
        )
        docs = ml_nlp.pipe(docs)
        docs = docs.set_processing(
            batch_size=2,
            num_gpu_workers=1,
            num_cpu_workers=1,
            gpu_worker_devices=["cpu"],
        )
        list(docs)
    assert "Deep learning error" in str(e.value)


@pytest.mark.parametrize(
    "backend",
    ["simple", "multiprocessing", "spark"],
)
def test_generator(backend):
    items = ["abc", "def", "ghij"]
    items = edsnlp.data.from_iterable(items)

    def gen(x):
        for char in x:
            yield char

    items = items.map(gen).set_processing(backend=backend, num_cpu_workers=2)
    # output from workers will be read in a round-robin fashion
    # ie zip(
    #   ("a",      "b",      "c",      "g", "h", "i", "j")  # worker 1
    #        ("d",      "e",      "f")  # worker 2
    # )
    assert set(items) == {"a", "d", "b", "e", "c", "f", "g", "h", "i", "j"}


@pytest.mark.parametrize("deterministic", [True, False])
def test_multiprocessing_sleep(deterministic):
    def process(x):
        if x % 2 == 0:
            time.sleep(0.1)
        return x

    items = list(range(100))
    items = edsnlp.data.from_iterable(items)
    items = items.map(process)
    items = items.set_processing(
        backend="multiprocessing",
        deterministic=deterministic,
        num_cpu_workers=2,
    )
    items = list(items)
    if deterministic:
        assert items == list(range(100))
    else:
        assert items != list(range(100))


@pytest.mark.parametrize("num_cpu_workers", [0, 1, 2])
def test_deterministic_skip(num_cpu_workers):
    def process_batch(x):
        return [i for i in x if i < 10 or i % 2 == 0]

    items = list(range(100))
    items = edsnlp.data.from_iterable(items)
    items = items.map_batches(process_batch)
    items = items.set_processing(
        deterministic=True,
        num_cpu_workers=num_cpu_workers,
    )
    items = list(items)
    assert items == [*range(0, 10), *range(10, 100, 2)]


@pytest.mark.parametrize(
    "backend",
    ["simple", "multiprocesing"],
)
@pytest.mark.skipif(torch is None, reason="torch not installed")
def test_backend_cache(backend):
    import torch

    from edsnlp.core.torch_component import (
        BatchInput,
        BatchOutput,
        TorchComponent,
        _caches,
    )

    @validate_arguments
    class InnerComponent(TorchComponent):
        def __init__(self, nlp=None, *args, **kwargs):
            super().__init__()
            self.called_forward = False

        def preprocess(self, doc):
            return {"text": doc.text}

        def collate(self, batch: Dict[str, Any]) -> BatchInput:
            return {"sizes": torch.as_tensor([len(x) for x in batch["text"]])}

        def forward(self, batch):
            assert not self.called_forward
            self.called_forward = True
            return {"sizes": batch["sizes"] * 2}

    @validate_arguments
    class OuterComponent(TorchComponent):
        def __init__(self, inner):
            super().__init__()
            self.inner = inner

        def preprocess(self, doc):
            return {"inner": self.inner.preprocess(doc)}

        def collate(self, batch: Dict[str, Any]) -> BatchInput:
            return {"inner": self.inner.collate(batch["inner"])}

        def forward(self, batch: BatchInput) -> BatchOutput:
            return {"inner": self.inner(batch["inner"])["sizes"].clone()}

        def postprocess(
            self,
            docs: Sequence[Doc],
            results: BatchOutput,
            inputs: List[Dict[str, Any]],
        ) -> Sequence[Doc]:
            return docs

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(InnerComponent(), name="inner")
    nlp.add_pipe(OuterComponent(nlp.pipes.inner), name="outer")
    text1 = "Word"
    text2 = "A phrase"
    text3 = "This is a sentence"
    text4 = "This is a longer document with many words."
    stream = edsnlp.data.from_iterable([text1, text2, text3, text4])
    stream = stream.map_pipeline(nlp)
    if backend == "simple":
        assert list(_caches) == []
        list(stream.set_processing(batch_size=4))
        assert list(_caches) == []
    elif backend == "multiprocessing":
        list(
            stream.set_processing(
                batch_size=2,
                num_gpu_workers=2,
                num_cpu_workers=1,
                gpu_worker_devices=["cpu", "cpu"],
            )
        )
    elif backend == "spark":
        list(stream.set_processing(backend="spark", batch_size=4))


def test_task_dispatch_schedule():
    fn = get_dispatch_schedule

    assert fn(0, range(4), range(2)) == [0, 0]
    assert fn(1, range(4), range(2)) == [1, 1]
    assert fn(2, range(4), range(2)) == [0, 0]
    assert fn(3, range(4), range(2)) == [1, 1]

    assert fn(0, range(3), range(2)) == [0, 0]
    assert fn(1, range(3), range(2)) == [1, 1]
    assert fn(2, range(3), range(2)) == [0, 1]

    assert fn(0, range(2), range(3)) == [0, 0, 2]
    assert fn(1, range(2), range(3)) == [1, 1, 2]
    assert fn(0, range(2), range(3)) == [0, 0, 2]
    assert fn(1, range(2), range(3)) == [1, 1, 2]

    assert fn(0, range(16), range(10)) == [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    assert fn(1, range(16), range(10)) == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    assert fn(2, range(16), range(10)) == [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    assert fn(3, range(16), range(10)) == [3, 3, 3, 3, 3, 3, 3, 3, 3, 3]
    assert fn(4, range(16), range(10)) == [4, 4, 4, 4, 4, 4, 4, 4, 4, 4]
    assert fn(5, range(16), range(10)) == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
    assert fn(6, range(16), range(10)) == [6, 6, 6, 6, 6, 6, 6, 6, 6, 6]
    assert fn(7, range(16), range(10)) == [7, 7, 7, 7, 7, 7, 7, 7, 7, 7]
    assert fn(8, range(16), range(10)) == [8, 8, 8, 8, 8, 8, 8, 8, 8, 8]
    assert fn(9, range(16), range(10)) == [9, 9, 9, 9, 9, 9, 9, 9, 9, 9]
    assert fn(10, range(16), range(10)) == [0, 0, 0, 0, 0, 0, 6, 6, 6, 6]
    assert fn(11, range(16), range(10)) == [1, 1, 1, 1, 1, 1, 7, 7, 7, 7]
    assert fn(12, range(16), range(10)) == [2, 2, 2, 2, 2, 2, 8, 8, 8, 8]
    assert fn(13, range(16), range(10)) == [3, 3, 3, 3, 3, 3, 9, 9, 9, 9]
    assert fn(14, range(16), range(10)) == [4, 4, 4, 4, 4, 4, 6, 7, 6, 7]
    assert fn(15, range(16), range(10)) == [5, 5, 5, 5, 5, 5, 8, 9, 8, 9]


def test_multiprocessing_on_simple_iterable_in_main():
    exec(
        """
import edsnlp.data

counter = 0

def complex_func(n):
    global counter
    counter += 1
    return n * n

stream = edsnlp.data.from_iterable(range(20))
stream = stream.map(complex_func)
stream = stream.set_processing(num_cpu_workers=2)
res = list(stream)
""",
        {"__MODULE__": "__main__"},
    )


def test_multiprocessing_on_full_example_in_main():
    exec(
        """
from spacy.tokens import Doc

import edsnlp
import edsnlp.pipes as eds
from edsnlp.data.converters import get_current_tokenizer

if not Doc.has_extension("note_text"):
    Doc.set_extension("note_text", default=None)
if not Doc.has_extension("date"):
    Doc.set_extension("date", default=None)
if not Doc.has_extension("person_id"):
    Doc.set_extension("person_id", default=None)


def convert_row_to_doc(row):
    if row["note_text"] is None:
        row["note_text"] = ""
    text = row["note_text"]
    doc = get_current_tokenizer()(text)
    doc._.note_id = row["note_id"]
    return doc


def convert_doc_to_row(doc_):
    note_id = doc_._.note_id
    person_id = doc_._.person_id
    note_text = doc_.text
    result = []
    for date in doc_.spans["dates"]:
        result.append(
            {
                "note_id": note_id,
                "person_id": person_id,
                # "note_text" : note_text,
                # "note_doc" : doc_,
                "date": date._.date.datetime,
            }
        )
    return result


nlp = edsnlp.blank("eds")
# nlp = eds_biomedic_aphp.load()
# nlp.add_pipe(eds.sections())
nlp.add_pipe(eds.dates())
nlp.add_pipe(eds.sentences())
data = edsnlp.data.from_iterable(
    [{"note_text": "Test", "note_id": "test"}],
    converter=convert_row_to_doc,
)
data = data.map_pipeline(nlp)
data_pd = data.set_processing(show_progress=True, num_cpu_workers=5).to_pandas(
    converter=convert_doc_to_row
)
""",
        {"__MODULE__": "__main__"},
    )



import string
from typing import List, cast

import regex
from spacy.tokens import Span

from edsnlp import registry
from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import BaseComponent
from edsnlp.pipes.misc.dates.models import AbsoluteDate
from edsnlp.utils.span_getters import SpanGetterArg, get_spans, validate_span_getter

noun_regex = r"""(?xi)
# No letter before
(?<![[:alpha:]])
(?:
    # Either a day of week
    (?P<day_of_week>
        lundi|
        mardi|
        mercredi|
        jeudi|
        vendredi|
        samedi|
        dimanche
    )|
    # Or a month
    (?P<month_1>jan[[:alpha:]]*)|
    (?P<month_2>fev[[:alpha:]]*|fév[[:alpha:]]*)|
    (?P<month_3>mar[[:alpha:]]*)|
    (?P<month_4>avr[[:alpha:]]*)|
    (?P<month_5>mai)|
    (?P<month_6>jun[[:alpha:]]*|juin[[:alpha:]]*)|
    (?P<month_7>jul[[:alpha:]]*|juil[[:alpha:]]*)|
    (?P<month_8>aou[[:alpha:]]*|aoû[[:alpha:]]*)|
    (?P<month_9>sep[[:alpha:]]*)|
    (?P<month_10>oct[[:alpha:]]*)|
    (?P<month_11>nov[[:alpha:]]*)|
    (?P<month_12>dec[[:alpha:]]*|déc[[:alpha:]]*)
)
"""

# no delimiter with space between numbers, ex: d d m m y y y y
full_regex = r"""(?x)
^
    (?P<year>(?:19|2[0Û])[\dÛ][\dÛ])
    [.-/](?P<month>[0Û][1-9]|1[Û012])
    [.-/](?P<day>[0Û][1-9]|[12][\dÛ]|3[Û01])
|
    (?P<day>[0Û][1-9]|[12][\dÛ]|3[Û01])
    [.-/]?(?P<month>[0Û][1-9]|1[Û012])
    [.-/]?(?P<year>(?:19|2[0Û])?[\dÛ][\dÛ])
|
    (?P<day>[0Û]\ +[1-9]|[12]\ +[\dÛ]|3\ +[Û01])
    \ *[.-/]?\ *(?<!\d)(?P<month>[0Û]\ +[1-9]|1\ +[Û012])
    \ *[.-/]?\ *(?<!\d)(?P<year>(?:1\ +9|2\ +[0Û])?\ +[\dÛ]\ +[\dÛ])
"""

number_regex = r"""(?x)
# Numbers with digits
[Û\d]+|
# Numbers with letters: we will sum the value of the matched groups (_{value})
(?i:
    (?<![[:alpha:]])
    (?:
        (?P<_1900>mil[[:alpha:]]*[-\s]?neuf[-\s]?cents?)|
        (?P<_2000>deux[-\s]?mil[[:alpha:]]*)
    )?[-\s]?
    (?:
        (?P<_20>vingt-?\s*et|vingt)|
        (?P<_30>trente-?\s*et|trente)|
        (?P<_40>quarante-?\s*et|quarante)|
        (?P<_50>cinquante-?\s*et|cinquante)|
        (?P<_60>soixante-?\s*et|soixante)|
        (?P<_80>quatres?-?\s*vingts?-?\s*et|quatres?-?\s*vingts?)
    )?[-\s]?
    (?:
        (?P<_1>premier|1\s*er|un)|
        (?P<_2>deux)|
        (?P<_3>trois)|
        (?P<_4>quatre)|
        (?P<_5>cinq)|
        (?P<_6>six)|
        (?P<_7>sept)|
        (?P<_8>huit)|
        (?P<_9>neuf)|
        (?P<_17>dix-?\s*sept)|
        (?P<_18>dix-?\s*huit)|
        (?P<_19>dix-?\s*neuf)|
        (?P<_10>dix)|
        (?P<_11>onze)|
        (?P<_12>douze)|
        (?P<_13>treize)|
        (?P<_14>quatorze)|
        (?P<_15>quinze)|
        (?P<_16>seize)
    )
)
"""


@registry.factory.register(
    "eds.dates_normalizer",
    requires=["doc.ents", "doc.spans"],
    assigns=["token._.date", "token._.date_format"],
)
class DatesNormalizer(BaseComponent):
    def __init__(
        self,
        nlp: PipelineProtocol,
        name: str = "dates_normalizer",
        span_getter: SpanGetterArg = {"ents": True},
    ):
        super().__init__(nlp=nlp, name=name)
        self.span_getter = validate_span_getter(span_getter)

    def set_extensions(self):
        if not Span.has_extension("date"):
            Span.set_extension("date", default=None)
        if not Span.has_extension("datetime"):
            Span.set_extension("datetime", default=None)
        if not Span.has_extension("date_format"):
            Span.set_extension("date_format", default=None)

    @staticmethod
    def extract_date(s, next_date=None, next_offsets=None):
        date_conf = {}

        m = regex.search(full_regex, s)
        if m:
            date = AbsoluteDate(
                day=int(m.group("day").replace(" ", "").replace("Û", "0")),
                month=int(m.group("month").replace(" ", "").replace("Û", "0")),
                year=int(m.group("year").replace(" ", "").replace("Û", "0")),
            )
            return date, sorted(
                [
                    (m.start("day"), m.end("day"), "d"),
                    (m.start("month"), m.end("month"), "m"),
                    (m.start("year"), m.end("year"), "y"),
                ]
            )

        matches = []
        remaining = []

        for match in regex.finditer(noun_regex, s):
            if match.group("day_of_week"):
                remaining.append((match.start(), match.end(), "w", None))
            else:
                value = next(m for m, value in match.groupdict().items() if value)
                matches.append((match.start(), match.end(), "m", int(value[6:])))

        numbers = list(regex.finditer(number_regex, s))
        for match in numbers:
            snippet = match.group()
            if snippet.replace("Û", "0").isdigit():
                value = int(snippet.replace("Û", "0"))
            else:
                value = sum(int(m[1:]) for m, v in match.groupdict().items() if v)
            if value == 0:
                matches.append((match.start(), match.end(), ".", None))
            elif value <= 12:
                if len(snippet) == 1:
                    matches.append((match.start(), match.end(), "dm", value))
                else:
                    matches.append((match.start(), match.end(), "dmy", value))
            elif value <= 31:
                matches.append((match.start(), match.end(), "dy", value))
            elif value <= 40:
                matches.append((match.start(), match.end(), "y", 2000 + value))
            elif 40 < value < 100:
                matches.append((match.start(), match.end(), "y", 1900 + value))
            elif 1900 <= value <= 2100:
                matches.append((match.start(), match.end(), "y", value))
            elif 1900 <= int(snippet[:4]) <= 2100:
                value = int(snippet[:4])
                matches.append((match.start(), match.start() + 4, "y", value))
            else:
                matches.append((match.start(), match.start() + 4, ".", value))

        matches = sorted(matches)

        pattern: List[str] = [m[2] for m in matches]  # type: ignore

        last = -1
        found = "".join(p for p in pattern if len(p) == 1)
        while len(found) != last:
            pattern = [(p.strip(found)) if len(p) > 1 else p for p in pattern]
            last = len(found)
            found = "".join(p for p in pattern if len(p) == 1)

        # In France, dates follow the d/m/y order
        if len(pattern) >= 2 and "d" in pattern[0] and "m" in pattern[1]:
            remaining = ["." if p in ("d", "m") else p for p in remaining]
            pattern = ["." if p in ("d", "m") else p for p in pattern]
            pattern[0] = "d"
            pattern[1] = "m"
        if len(pattern) >= 2 and "m" in pattern[-2] and "y" in pattern[-1]:
            remaining = ["." if p in ("m", "y") else p for p in remaining]
            pattern = ["." if p in ("m", "y") else p for p in pattern]
            pattern[-2] = "m"
            pattern[-1] = "y"
        if len(pattern) >= 3 and "y" in pattern[0] and "m" in pattern[1]:
            remaining = ["." if p in ("d", "m", "y") else p for p in remaining]
            pattern = ["." if p in ("d", "m", "y") else p for p in pattern]
            pattern[0] = "y"
            pattern[1] = "m"
            pattern[2] = "d"

        # Handle cases like [10] in "du [10] au [12/08/1995]"
        if (
            pattern
            and next_date is not None
            and "m" not in pattern
            and "y" not in pattern
            and len(["d" in p for p in pattern]) == 1
            and next_date.month is not None
            and next_date.day is not None
            and next_offsets[0][2] not in "ym"
        ):
            if next_date.year is not None:
                date_conf["year"] = next_date.year
            date_conf["month"] = next_date.month
            found += "ym"
            pattern = ["d" if "d" in p else p for p in pattern]
        # Handle cases like [10] in "de [mai] à [juin 1996]"
        elif (
            pattern
            and next_date is not None
            and "y" not in pattern
            and len(["m" in p for p in pattern]) == 1
            and next_date.year is not None
            and next_date.month is not None
            and next_offsets[0][2] not in "yd"
        ):
            date_conf["year"] = next_date.year
            found += "y"
            pattern = ["m" if "m" in p else p for p in pattern]

        # Handle year found but missing month => remove day
        forbidden = ""
        found = pattern + [m[2] for m in remaining]
        if "d" in found and "y" in found and "m" not in found:
            forbidden += "d"

        if len(["d" in p for p in pattern]) == 1:
            pattern = ["d" if "d" in p else p for p in pattern]

        # Handle missing day => remove day of week
        if "w" in found and "d" not in found:
            forbidden += "w"

        matches = sorted(
            [
                (m[0], m[1], p if len(p) == 1 and p not in forbidden else ".", m[3])
                for m, p in (
                    *zip(matches, pattern),
                    *zip(remaining, [m[2] for m in remaining]),
                )
            ]
        )

        for m in matches:
            if m[2] == "d":
                date_conf["day"] = m[3]
            elif m[2] == "m":
                date_conf["month"] = m[3]
            elif m[2] == "y":
                date_conf["year"] = m[3]

        date = cast(AbsoluteDate, date_conf)

        return date, [(b, e, k) for b, e, k, v in matches]

    def extract_format(self, s, matches):
        date_format = ""
        offset = 0
        translation = str.maketrans(string.digits, "?" * len(string.digits))
        for begin, end, kind in matches:
            date_format += s[offset:begin].translate(translation).replace("%", "%%")
            snippet = s[begin:end]
            if kind == "d":
                if snippet.isdigit() and len(snippet) == 2:
                    date_format += "%d"
                else:
                    date_format += "%-d"
            elif kind == "m":
                if snippet.isdigit():
                    if len(snippet) == 1:
                        date_format += "%-m"
                    else:
                        date_format += "%m"
                else:
                    if len(snippet) <= 3:
                        date_format += "%b"
                    else:
                        date_format += "%B"
            elif kind == "w":
                if len(snippet) <= 3:
                    date_format += "%a"
                else:
                    date_format += "%A"
            elif kind == "y":
                date_format += "%Y"
            elif kind == ".":
                date_format += snippet.translate(translation).replace("%", "%%")
            offset = end

        date_format += s[offset:].translate(translation).replace("%", "%%")

        return date_format

    def __call__(self, doc):
        spans = list(get_spans(doc, self.span_getter))
        last_date = None
        last_date_offsets = None
        last_span = None
        for span in reversed(spans):
            span: Span
            text = span.text
            if last_span and last_span.start - span.end > 2:
                last_date = last_date_offsets = None

            date, date_offsets = self.extract_date(
                text,
                last_date,
                last_date_offsets,
            )
            span._.date = date
            span._.datetime = date.to_datetime(doc._.note_datetime)
            span._.date_format = self.extract_format(text, date_offsets)

            last_span, last_date, last_date_offsets = span, date, date_offsets

        return doc



from typing import Union

import numpy as np
import pytest
from spacy.tokens import Doc

import edsnlp
from edsnlp.utils.fuzzy_alignment import AnnotatedText, align, regex_multisub_with_spans


def test_regex_multi_sub_with_spans():
    kw = dict(
        patterns=["===", r"hello (world)", "example"],
        replacements=[" ", r"hi \1", "sample"],
        text="=== hello world! This is a hello world example.===",
    )
    new_text, deltas = regex_multisub_with_spans(**kw)
    assert new_text == "  hi world! This is a hi world sample. "
    new_text, deltas = regex_multisub_with_spans(**kw, return_deltas=True)
    assert new_text == "  hi world! This is a hi world sample. "
    # entities=[
    #     {"label": "GREET", "fragments": [{"begin": 4, "end": 9}]},
    #     {"label": "WORLD", "fragments": [{"begin": 10, "end": 16}]},
    #     {"label": "EXAMPLE", "fragments": [{"begin": 39, "end": 46}]},
    # ]
    new_begins = deltas.apply(np.asarray([4, 10, 39]), side="left")
    new_ends = deltas.apply(np.asarray([9, 16, 46]), side="right")
    assert (new_begins == np.asarray([2, 2, 31])).all()
    # since we replace "hello world" by "hi world", the
    # span "hello" becomes "hi world", hence 10 instead of 4
    assert (new_ends == np.asarray([10, 11, 37])).all()

    old_begins = deltas.unapply(np.asarray([2, 5, 31, 33]), side="left")
    old_ends = deltas.unapply(np.asarray([10, 11, 37, 35]), side="right")
    assert (old_begins == np.asarray([4, 4, 39, 39])).all()
    assert (old_ends == np.asarray([15, 16, 46, 46])).all()


def conv_from_doc(doc: Union[Doc, str]):
    if isinstance(doc, str):
        return {"doc_id": None, "text": doc, "entities": []}
    return {
        "doc_id": doc._.note_id,
        "text": doc.text,
        "entities": [
            {
                "label": ent.label_,
                "fragments": [{"begin": ent.start_char, "end": ent.end_char}],
            }
            for ent in doc.ents
        ],
    }


def conv_to_doc(annotated: AnnotatedText, tok=None) -> Doc:
    if tok is None:
        tok = edsnlp.data.converters.get_current_tokenizer()
    doc = tok(annotated["text"])
    ents = []
    for ent in annotated["entities"]:
        for frag in ent["fragments"]:
            span = doc.char_span(
                frag["begin"], frag["end"], label=ent["label"], alignment_mode="expand"
            )
            if span is not None:
                ents.append(span)
    doc.ents = ents
    doc._.note_id = annotated["doc_id"]
    return doc


@pytest.mark.parametrize("high_threshold", [True, False])
def test_align(high_threshold, md2doc, doc2md):
    old = (
        "This is a [small sample](SAMPLE)\n\n[\ndoc\n](DOC)  . "
        "It contains some [texts](TEXT) to be annotated. "
    )
    new = (
        "This is a  modified  [small   sample](SAMPLE) [\ndoc\n](DOC) . "
        "It contains some [text](TEXT) to be annotated."
    )
    if not high_threshold:
        res = align(
            conv_from_doc(md2doc(old)),
            conv_from_doc(md2doc(new)),
            do_debug=True,
        )
        result = conv_to_doc(res["doc"])
        assert doc2md(result) == new
    else:
        res = align(
            conv_from_doc(md2doc(old)),
            conv_from_doc(md2doc(new)),
            do_debug=True,
            threshold=20,
        )
        result = conv_to_doc(res["doc"])
        assert doc2md(result) == md2doc(new).text


def test_ambiguous_exact_match(md2doc, doc2md):
    old = (
        "this is ambiguous: xy , where should we annotate ?\n" * 25
        + "this is ambiguous [xy](AMBIGUOUS) , where should we annotate ?\n"
        + "this is ambiguous: xy , where should we annotate ?\n" * 24
    )
    new = "this is ambiguous: xy , where should we annotate ?\n" * 50
    res = align(
        conv_from_doc(md2doc(old)),
        conv_from_doc(md2doc(new)),
        do_debug=True,
        sim_scheme=[(10, 0.7)],
    )
    result = conv_to_doc(res["doc"])
    assert doc2md(result) == new


def test_ambiguous_inexact_match(md2doc, doc2md):
    old = (
        "this is ambiguous: xy , where should we annotate ?\n" * 25
        + "this is ambiguous [xz](AMBIGUOUS) , where should we annotate ?\n"
        + "this is ambiguous: xy , where should we annotate ?\n" * 24
    )
    new = "this is ambiguous: xy , where should we annotate ?\n" * 50
    res = align(
        conv_from_doc(md2doc(old)),
        conv_from_doc(md2doc(new)),
        do_debug=True,
        sim_scheme=[(10, 0.7)],
    )
    result = conv_to_doc(res["doc"])
    assert doc2md(result) == new


def test_missing(md2doc, doc2md):
    old = "this is a sample: [abcd](MISSING), where[](EMPTY) should we annotate ?"
    new = "the cat sat on the mat, and eat mice when he is hungry."
    res = align(
        conv_from_doc(md2doc(old)),
        conv_from_doc(md2doc(new)),
        do_debug=True,
        sim_scheme=[(10, 0.7)],
    )
    result = conv_to_doc(res["doc"])
    assert doc2md(result) == new



from typing import TYPE_CHECKING

from edsnlp import registry

from .extractive_qa import TrainableExtractiveQA

create_component = registry.factory.register(
    "eds.extractive_qa",
    assigns=[],
    deprecated=[],
)(TrainableExtractiveQA)

if TYPE_CHECKING:
    create_component = TrainableExtractiveQA



import re
from typing import Dict, Optional, Union

import pandas as pd
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.phrase import EDSPhraseMatcher
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.pipes.base import BaseComponent
from edsnlp.pipes.misc.tables import patterns
from edsnlp.utils.typing import AsList


class TablesMatcher(BaseComponent):
    '''
    The `eds.tables` matcher detects tables in a documents.

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.tables())

    text = """
    SERVICE
    MEDECINE INTENSIVE –
    REANIMATION
    Réanimation / Surveillance Continue
    Médicale

    COMPTE RENDU D'HOSPITALISATION du 05/06/2020 au 10/06/2020
    Madame DUPONT Marie, née le 16/05/1900, âgée de 20 ans, a été hospitalisée en
    réanimation du 05/06/1920 au 10/06/1920 pour intoxication médicamenteuse volontaire.

    Examens complémentaires
    Hématologie
    Numération
    Leucocytes ¦x10*9/L ¦4.97 ¦4.09-11
    Hématies ¦x10*12/L¦4.68 ¦4.53-5.79
    Hémoglobine ¦g/dL ¦14.8 ¦13.4-16.7
    Hématocrite ¦% ¦44.2 ¦39.2-48.6
    VGM ¦fL ¦94.4 + ¦79.6-94
    TCMH ¦pg ¦31.6 ¦27.3-32.8
    CCMH ¦g/dL ¦33.5 ¦32.4-36.3
    Plaquettes ¦x10*9/L ¦191 ¦172-398
    VMP ¦fL ¦11.5 + ¦7.4-10.8

    Sur le plan neurologique : Devant la persistance d'une confusion à distance de
    l'intoxication au
    ...

    2/2Pat : <NOM> <Prenom>|F |<date> | <ipp> |Intitulé RCP
    """

    doc = nlp(text)

    # A table span
    table = doc.spans["tables"][0]

    # Leucocytes ¦x10*9/L ¦4.97 ¦4.09-11
    # Hématies ¦x10*12/L¦4.68 ¦4.53-5.79
    # Hémoglobine ¦g/dL ¦14.8 ¦13.4-16.7
    # Hématocrite ¦% ¦44.2 ¦39.2-48.6
    # VGM ¦fL ¦94.4 + ¦79.6-94
    # TCMH ¦pg ¦31.6 ¦27.3-32.8
    # CCMH ¦g/dL ¦33.5 ¦32.4-36.3
    # Plaquettes ¦x10*9/L ¦191 ¦172-398
    # VMP ¦fL ¦11.5 + ¦7.4-10.8

    # Convert span to Pandas table
    df = table._.to_pd_table(
        as_spans=False,  # set True to set the table cells as spans instead of strings
        header=False,  # set True to use the first row as header
        index=False,  # set True to use the first column as index
    )
    type(df)
    # Out: <class 'pandas.core.frame.DataFrame'>
    ```
    The pandas DataFrame:

    |      | 0           | 1        | 2      | 3         |
    | ---: | :---------- | :------- | :----- | :-------- |
    |    0 | Leucocytes  | x10*9/L  | 4.97   | 4.09-11   |
    |    1 | Hématies    | x10*12/L | 4.68   | 4.53-5.79 |
    |    2 | Hémoglobine | g/dL     | 14.8   | 13.4-16.7 |
    |    3 | Hématocrite | %        | 44.2   | 39.2-48.6 |
    |    4 | VGM         | fL       | 94.4 + | 79.6-94   |
    |    5 | TCMH        | pg       | 31.6   | 27.3-32.8 |
    |    6 | CCMH        | g/dL     | 33.5   | 32.4-36.3 |
    |    7 | Plaquettes  | x10*9/L  | 191    | 172-398   |
    |    8 | VMP         | fL       | 11.5 + | 7.4-10.8  |

    Extensions
    ----------
    The `eds.tables` pipeline declares the `span._.to_pd_table()` Span extension.
    This function returns a parsed pandas version of the table.

    Parameters
    ----------
    nlp : PipelineProtocol
        Pipeline object
    name: str
        Name of the component.
    tables_pattern : Optional[Dict[str, str]]
        The regex pattern to identify tables.
        The key of dictionary should be `tables`
    sep_pattern : Optional[str]
        The regex pattern to identify the separator pattern.
        Used when calling `to_pd_table`.
    min_rows : Optional[int]
        Only tables with more then `min_rows` lines will be detected.
    attr : str
        spaCy's attribute to use:
        a string with the value "TEXT" or "NORM", or a dict with
        the key 'term_attr'. We can also add a key for each regex.
    ignore_excluded : bool
        Whether to skip excluded tokens.

    Authors and citation
    --------------------
    The `eds.tables` pipeline was developed by AP-HP's Data Science team.
    '''

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "tables",
        *,
        tables_pattern: Optional[AsList[str]] = None,
        sep_pattern: Optional[AsList[str]] = None,
        min_rows: int = 2,
        attr: Union[Dict[str, str], str] = "TEXT",
        ignore_excluded: bool = True,
    ):
        super().__init__(nlp, name)
        if tables_pattern is None:
            tables_pattern = patterns.regex_template

        if sep_pattern is None:
            sep_pattern = patterns.sep

        self.regex_matcher = RegexMatcher(attr=attr, ignore_excluded=ignore_excluded)
        self.regex_matcher.add(
            "table",
            list(
                dict.fromkeys(
                    template.format(sep=re.escape(sep), n=re.escape(str(min_rows)))
                    for sep in sep_pattern
                    for template in tables_pattern
                )
            ),
        )

        self.term_matcher = EDSPhraseMatcher(
            nlp.vocab, attr=attr, ignore_excluded=ignore_excluded
        )
        self.term_matcher.build_patterns(
            nlp,
            {
                "eol_pattern": "\n",
                "sep_pattern": sep_pattern,
            },
        )

        if not Span.has_extension("to_pd_table"):
            Span.set_extension("to_pd_table", method=self.to_pd_table)

    @classmethod
    def set_extensions(cls) -> None:
        """
        Set extensions for the tables pipeline.
        """

        if not Span.has_extension("table"):
            Span.set_extension("table", default=None)

    def get_table(self, table):
        """
        Convert spans of tables to dictionaries
        Parameters
        ----------
        table : Span

        Returns
        -------
        List[Span]
        """

        # We store each row in a list and store each of hese lists
        # in processed_table for post processing
        # considering the self.col_names and self.row_names var
        processed_table = []
        delimiters = [
            delimiter
            for delimiter in self.term_matcher(table, as_spans=True)
            if delimiter.start >= table.start and delimiter.end <= table.end
        ]

        last = table.start
        row = []
        # Parse the table to match each cell thanks to delimiters
        for delimiter in delimiters:
            row.append(table[last - table.start : delimiter.start - table.start])
            last = delimiter.end

            # End the actual row if there is an end of line
            if delimiter.label_ == "eol_pattern":
                processed_table.append(row)
                row = []

        # Remove first or last column in case the separator pattern is
        # also used in the raw table to draw the outlines
        max_len = max(len(row) for row in processed_table)
        if all(row[0].start == row[0].end for row in processed_table):
            processed_table = [row[1:] for row in processed_table]
        if all(
            row[-1].start == row[-1].end
            for row in processed_table
            if len(row) == max_len
        ):
            processed_table = [row[:-1] for row in processed_table]

        return processed_table

    def __call__(self, doc: Doc) -> Doc:
        """
        Find spans that contain tables

        Parameters
        ----------
        doc : Doc

        Returns
        -------
        Doc
        """
        matches = list(self.regex_matcher(doc, as_spans=True))
        doc.spans["tables"] = matches
        return doc

    def to_pd_table(
        self,
        span,
        as_spans=False,
        header: bool = False,
        index: bool = False,
    ) -> pd.DataFrame:
        """
        Return pandas DataFrame

        Parameters
        ----------
        span : Span
            The span containing the table
        as_spans : bool
            Whether to return the table cells as spans
        header : bool
            Whether the table has a header
        index : bool
            Whether the table has an index
        """
        table = self.get_table(span)
        if not as_spans:
            table = [[str(cell) for cell in data] for data in table]

        table = pd.DataFrame.from_records(table)
        if header:
            table.columns = [str(k) for k in table.iloc[0]]
            table = table[1:]
        if index:
            table.index = [str(k) for k in table.iloc[:, 0]]
            table = table.iloc[:, 1:]
        return table



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Tuple,
    TypeVar,
    Union,
)

from edsnlp.utils.span_getters import SeqStr, SpanFilter
from edsnlp.utils.typing import Validated

Binding = Tuple[str, Any]


def _check_path(path: str):
    assert [letter.isalnum() or letter == "_" or letter == "." for letter in path], (
        "The label must be a path of valid python identifier to be used as a getter"
        "in the following template: span.[YOUR_LABEL], such as `label_` or `_.negated"
    )
    parts = path.split(".")
    new_path = "span"
    for part in parts:
        if " " in part:
            new_path = "getattr(" + new_path + f", {part!r})"
        elif len(part) > 0:
            new_path += "." + part
    return new_path


def make_binding_getter(attribute: Union[str, Binding]):
    """
    Make a attribute getter

    Parameters
    ----------
    attribute: Union[str, Binding]
        Either one of the following:
        - a path to a nested attributes of the span, such as "attribute_" or "_.negated"
        - a tuple of (key, value) equality, such as `("_.date.mode", "PASSED")`

    Returns
    -------
    Callable[[Span], bool]
        The attribute getter
    """
    if isinstance(attribute, tuple):
        path, value = attribute
        path = _check_path(path)
        ctx = {"value": value}
        exec(
            f"def getter(span):\n"
            f"    try:\n"
            f"        return {path} == value\n"
            f"    except (AttributeError, KeyError):\n"
            f"        return False\n",
            ctx,
            ctx,
        )
        return ctx["getter"]
    else:
        path = _check_path(attribute)
        ctx = {}
        exec(
            f"def getter(span):\n"
            f"    try:\n"
            f"        return {path}\n"
            f"    except (AttributeError, KeyError):\n"
            f"        return None\n",
            ctx,
            ctx,
        )
        return ctx["getter"]


def make_binding_setter(binding: Binding):
    """
    Make a attribute setter

    Parameters
    ----------
    binding: Binding
        A pair of
        - a path to a nested attributes of the span, such as `attribute_` or `_.negated`
        - a value assignment

    Returns
    -------
    Callable[[Span]]
        The attribute setter
    """
    if isinstance(binding, tuple):
        path, value = binding
        path = _check_path(path)
        fn_string = f"""def setter(span): {path} = value"""
        ctx = {"value": value}
        exec(fn_string, ctx, ctx)
    else:
        path = _check_path(binding)
        fn_string = f"""def setter(span, value): {path} = value"""
        ctx = {}
        exec(fn_string, ctx, ctx)
    return ctx["setter"]


K = TypeVar("K")
V = TypeVar("V")


class keydefaultdict(dict):
    def __init__(self, default_factory: Callable[[K], V]):
        super().__init__()
        self.default_factory = default_factory

    def __missing__(self, key: K) -> V:
        ret = self[key] = self.default_factory(key)
        return ret


BINDING_GETTERS = keydefaultdict(make_binding_getter)
BINDING_SETTERS = keydefaultdict(make_binding_setter)

Attributes = Dict[str, SpanFilter]


def validate_attributes(value: Union[SeqStr, Dict[str, SpanFilter]]) -> Attributes:
    if callable(value):
        return value
    if isinstance(value, str):
        return {value: True}
    if isinstance(value, list):
        return {qlf: True for qlf in value}
    elif isinstance(value, dict):
        new_value = {}
        for k, v in value.items():
            if isinstance(v, bool):
                new_value[k] = v
            elif isinstance(v, str):
                new_value[k] = [v]
            elif isinstance(v, (list, tuple)) and all(isinstance(i, str) for i in v):
                new_value[k] = list(v)
            else:
                raise TypeError(
                    f"Invalid entry {value} ({type(value)}) for Attributes, "
                    f"expected bool/string(s), dict of bool/string(s) or callable"
                )
        return new_value
    else:
        raise TypeError(
            f"Invalid entry {value} ({type(value)}) for SpanSetterArg, "
            f"expected bool/string(s), dict of bool/string(s) or callable"
        )


class AttributesArg(Validated):
    """
    Valid values for the `attributes` argument of a component can be :

    - a (span) -> attribute callable
    - a attribute name ("_.negated")
    - a list of attribute names (["_.negated", "_.event"])
    - a dict of attribute name to True or list of labels, to filter the attributes

    Examples
    --------
    - `attributes="_.negated"` will use the `negated` extention of the span
    - `attributes=["_.negated", "_.past"]` will use the `negated` and `past`
       extensions of the span
    - `attributes={"_.negated": True, "_.past": "DATE"}` will use the `negated`
       extension of any span, and the `past` extension of spans with the `DATE` label
    """

    @classmethod
    def validate(cls, value, config=None) -> Attributes:
        return validate_attributes(value)


if TYPE_CHECKING:
    AttributesArg = Union[SeqStr, Dict[str, Union[bool, SeqStr]], Callable]  # noqa: F811



import functools
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Generic, List, TypeVar, Union

import pydantic
from confit import Validatable
from confit.errors import patch_errors

T = TypeVar("T")

if pydantic.VERSION >= "2":
    from pydantic_core import core_schema


Validated = Validatable


class MetaAsList(type):
    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        type_ = next((base.type_ for base in bases if hasattr(base, "type_")), Any)
        cls.type_ = type_

    @functools.lru_cache(maxsize=None)
    def __getitem__(self, item):
        new_type = MetaAsList(self.__name__, (self,), {})
        new_type.type_ = item
        return new_type

    def validate(cls, value, config=None):
        if isinstance(value, dict):
            value = [value]
        if not isinstance(value, list):
            value = [value]
        try:
            return cast(List[cls.type_], value)
        except pydantic.ValidationError as e:
            e = patch_errors(e, drop_names=("__root__",))
            e.model = cls
            raise e

    def __get_validators__(cls):
        yield cls.validate

    def __get_pydantic_core_schema__(cls, source, handler):
        return core_schema.no_info_plain_validator_function(cls.validate)


class AsList(Generic[T], metaclass=MetaAsList):
    pass


if pydantic.VERSION < "2":

    def cast(type_, obj):
        class Model(pydantic.BaseModel, arbitrary_types_allowed=True):
            __root__: type_

        return Model(__root__=obj).__root__
else:
    from dataclasses import is_dataclass

    from pydantic import BaseModel
    from pydantic.type_adapter import ConfigDict, TypeAdapter
    from typing_extensions import is_typeddict

    @lru_cache(maxsize=32)
    def make_type_adapter(type_):
        config = None

        if not (
            (isinstance(type_, type) and issubclass(type_, BaseModel))
            or is_dataclass(type_)
            or is_typeddict(type_)
        ):
            config = ConfigDict(arbitrary_types_allowed=True)
        return TypeAdapter(type_, config=config)

    def cast(type_, obj):
        return make_type_adapter(type_).validate_python(obj)


if TYPE_CHECKING:
    AsList = Union[T, List[T]]  # noqa: F811



"""
EDS-NLP
"""

import sys
import importlib.abc
import importlib.util
from pathlib import Path
from spacy import pipeline as _spacy_pipeline  # noqa: F401
from . import extensions
from . import patch_spacy
from .core.pipeline import Pipeline, blank, load
from .core.registries import registry
import edsnlp.data  # noqa: F401
import edsnlp.pipes
from . import reducers

__version__ = "0.19.0"

BASE_DIR = Path(__file__).parent


# Everything below is to support deprecated use of edsnlp.pipelines
# route imports of submodules of edsnlp.pipelines to their edsnlp.pipes counterparts.
# The same is done for edsnlp.scorers -> edsnlp.metrics

class AliasPathFinder(importlib.abc.MetaPathFinder):
    def find_spec(self, fullname, path, target=None):  # pragma: no cover
        if not fullname.startswith("edsnlp."):
            return None
        if fullname.startswith("edsnlp.pipelines"):
            new_name = "edsnlp.pipes" + fullname[16:]
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec
        if fullname.startswith("edsnlp.core.lazy_collection"):
            new_name = "edsnlp.core.stream" + fullname[27:]
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec
        if fullname.startswith("edsnlp.optimization"):
            new_name = "edsnlp.training.optimizer" + fullname[19:]
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec
        if fullname.startswith("edsnlp.scorers"):
            new_name = "edsnlp.metrics" + fullname[14:]
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec
        if fullname.startswith("edsnlp.metrics.span_classification"):
            new_name = "edsnlp.metrics.span_attribute" + fullname[34:]
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec
        if fullname.startswith("edsnlp.metrics.span_attributes"):
            new_name = "edsnlp.metrics.span_attribute" + fullname[30:]
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec
        if fullname == "edsnlp.utils.regex":
            new_name = "edsnlp.utils.regex_utils"
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec
        if "span_qualifier" in fullname.split("."):
            new_name = fullname.replace("span_qualifier", "span_classifier")
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec
        if "measurements" in fullname.split("."):
            new_name = fullname.replace("measurements", "quantities")
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec


class AliasLoader(importlib.abc.Loader):
    def __init__(self, alias):
        self.alias = alias

    def create_module(self, spec):
        # import the edsnlp.pipe.* module first, then alias it to edsnlp.pipelines.*
        module = importlib.import_module(self.alias)
        sys.modules[spec.name] = module
        return module

    def exec_module(self, module):
        pass


# Add the custom finder to sys.path_hooks
sys.meta_path.insert(0, AliasPathFinder())



from edsnlp.data.converters import MarkupToDocConverter
from edsnlp.pipes.misc.explode import Explode
from edsnlp.utils.span_getters import get_spans_with_group


def make_doc():
    """
    Build a single document with two entities (entity, date) in `doc.ents`
    and one adjective span in the custom span-group `adj`.
    """
    converter = MarkupToDocConverter(
        preset="xml",
        span_setter={"ents": ["entity", "date"], "adj": ["adj"]},
    )
    return converter(
        "Ceci est un <entity>texte</entity> très <adj>important</adj>, "
        "écrit le <date is_recent>25 juil. 2025</date>"
    )


def groups_in(doc):
    """
    Return the set of non-empty span-group names contained in `doc`.
    (`ents` is treated as a group of its own.)
    """
    groups = set()
    if len(doc.ents):
        groups.add("ents")
    groups.update(name for name, spans in doc.spans.items() if len(spans))
    return groups


# ---------------------------------------------------------------------------
# Unit tests
# ---------------------------------------------------------------------------
def test_explode_multiple_groups():
    """
    • Exploding on both ``ents`` and ``adj`` must yield exactly three docs.
    • Each exploded doc must keep the full text, but contain **one** and only
      one non-empty span-group.
    • The surviving span must be the right one.
    """
    doc = make_doc()
    exploder = Explode(span_getter=["ents", "adj"])

    exploded = list(exploder(doc))
    assert len(exploded) == 3

    # In any order, we expect the three following "profiles"
    profiles = sorted(
        [
            (
                tuple(e.text for e in d.ents),
                [e.text for e in d.spans.get("adj", [])],
            )
            for d in exploded
        ]
    )

    assert profiles == sorted(
        [
            (("texte",), []),
            (("25 juil. 2025",), []),
            ((), ["important"]),
        ]
    )

    # Check that attributes are correctly preserved
    assert exploded[1].ents[0]._.is_recent is True
    assert not exploded[0].ents[0]._.is_recent

    # Each sub-document must contain *exactly* one non-empty span-group
    assert all(len(groups_in(sub)) == 1 for sub in exploded)

    # Original doc remains unaltered
    assert groups_in(doc) == {"ents", "adj"}


def test_explode_single_group():
    """
    Exploding only on ``ents`` yields one doc per entity and leaves
    the ``adj`` group untouched in the originals.
    """
    doc = make_doc()
    exploder = Explode(span_getter="ents")

    exploded = list(exploder(doc))
    assert len(exploded) == 2
    assert all(len(sub.ents) == 1 for sub in exploded)
    assert sorted(e.text for sub in exploded for e in sub.ents) == [
        "25 juil. 2025",
        "texte",
    ]


def test_explode_filter_expr():
    doc = make_doc()
    exploder = Explode(
        span_getter=["ents", "adj"],
        filter_expr="any('t' in e.text for e in (*doc.ents, *doc.spans['adj']))",
    )

    exploded = list(exploder(doc))
    assert len(exploded) == 2
    ents = [
        (e.text, g)
        for sub in exploded
        for e, g in get_spans_with_group(sub, exploder.span_getter)
    ]
    assert ents == [("texte", "ents"), ("important", "adj")]


def test_explode_no_spans():
    """
    When the document has no spans selected by the getter, the original
    doc must be yielded unchanged (identity check included).
    """
    import spacy

    nlp = spacy.blank("fr")
    doc = nlp("Pas d'entités ici.")
    exploder = Explode(span_getter="ents")

    result = list(exploder(doc))
    assert result == [doc]  # same object
    assert len(result[0].ents) == 0



import inspect
import types
from contextlib import contextmanager
from dataclasses import dataclass
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
)
from weakref import WeakKeyDictionary

import catalogue
import spacy
from confit import Config, Registry, RegistryCollection, set_default_registry
from confit.errors import ConfitValidationError, patch_errors
from confit.registry import Draft
from spacy.pipe_analysis import validate_attrs

import edsnlp
from edsnlp.utils.collections import FrozenDict, FrozenList

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

PIPE_META = WeakKeyDictionary()


def accepted_arguments(
    func: Callable,
    args: Sequence[str],
) -> Sequence[str]:
    """
    Checks that a function accepts a list of keyword arguments

    Parameters
    ----------
    func: Callable[..., T]
        Function to check
    args: Union[str, Sequence[str]]
        Argument or list of arguments to check

    Returns
    -------
    List[str]
    """
    sig = inspect.signature(func)
    has_kwargs = any(
        param.kind == param.VAR_KEYWORD for param in sig.parameters.values()
    )
    if has_kwargs:
        return args
    return [arg for arg in args if arg in sig.parameters]


@dataclass
class FactoryMeta:
    assigns: Iterable[str]
    requires: Iterable[str]
    retokenizes: bool
    default_config: Dict


T = TypeVar("T")


_AUTO_INSTANTIATE_COMPLETE_PIPE_DRAFT = True


class DraftPipe(Draft[T]):
    def __init__(self, func, kwargs):
        super().__init__(func, kwargs)
        self.instantiated = None
        self.error = None

    @staticmethod
    @contextmanager
    def disable_auto_instantiation():
        global _AUTO_INSTANTIATE_COMPLETE_PIPE_DRAFT
        old_value = _AUTO_INSTANTIATE_COMPLETE_PIPE_DRAFT
        _AUTO_INSTANTIATE_COMPLETE_PIPE_DRAFT = False
        try:
            yield
        finally:
            _AUTO_INSTANTIATE_COMPLETE_PIPE_DRAFT = old_value

    def maybe_nlp(self) -> Union["DraftPipe", Any]:
        """
        If the factory requires an nlp argument and the user has explicitly
        provided it (this is unusual, we usually expect the factory to be
        instantiated via add_pipe, or a config), then we should instantiate
        it.

        Returns
        -------
        Union["PartialFactory", Any]
        """
        from edsnlp.core.pipeline import Pipeline, PipelineProtocol

        sig = inspect.signature(self._func)
        if (
            _AUTO_INSTANTIATE_COMPLETE_PIPE_DRAFT
            and (
                not (
                    "nlp" in sig.parameters
                    and (
                        sig.parameters["nlp"].default is sig.empty
                        or sig.parameters["nlp"].annotation
                        in (Pipeline, PipelineProtocol)
                    )
                )
                or "nlp" in self._kwargs
            )
            and not self.search_nested_drafts(self._kwargs)
        ):
            return self._func(**self._kwargs)
        return self

    @classmethod
    def search_nested_drafts(cls, obj):
        if isinstance(obj, DraftPipe):
            return obj
        elif isinstance(obj, dict):
            for value in obj.values():
                result = cls.search_nested_drafts(value)
                if result is not None:
                    return result
        elif isinstance(obj, (tuple, list, set)):
            for value in obj:
                result = cls.search_nested_drafts(value)
                if result is not None:
                    return result
        return None

    def instantiate(
        self,
        nlp: "edsnlp.Pipeline",
        path: Optional[Sequence[str]] = (),
    ):
        """
        To ensure compatibility with spaCy's API, we need to support
        passing in the nlp object and name to factories. Since they can be
        nested, we need to add them to every factory in the config.
        """
        if isinstance(self, DraftPipe):
            if self.error is not None:
                raise self.error

            if self.instantiated is not None:
                return self.instantiated

            name = path[0] if len(path) == 1 else None
            parameters = (
                inspect.signature(self._func.__init__).parameters
                if isinstance(self._func, type)
                else inspect.signature(self._func).parameters
            )
            kwargs = {
                key: DraftPipe.instantiate(
                    self=value,
                    nlp=nlp,
                    path=(*path, key),
                )
                for key, value in self._kwargs.items()
            }
            try:
                if nlp and "nlp" in parameters:
                    kwargs["nlp"] = nlp
                if name and "name" in parameters:
                    kwargs["name"] = name
                self.instantiated = self._func(**kwargs)
            except ConfitValidationError as e:
                self.error = e
                raise ConfitValidationError(
                    patch_errors(e.raw_errors, path, model=e.model),
                    model=e.model,
                    name=self._func.__module__ + "." + self._func.__qualname__,
                )  # .with_traceback(None)
            # except Exception as e:
            #     obj.error = e
            #     raise ConfitValidationError([ErrorWrapper(e, path)])
            return self.instantiated
        elif isinstance(self, dict):
            instantiated = {}
            errors = []
            for key, value in self.items():
                try:
                    instantiated[key] = DraftPipe.instantiate(
                        self=value,
                        nlp=nlp,
                        path=(*path, key),
                    )
                except ConfitValidationError as e:
                    errors.extend(e.raw_errors)
            if errors:
                raise ConfitValidationError(errors)
            return instantiated
        elif isinstance(self, (tuple, list)):
            instantiated = []
            errors = []
            for i, value in enumerate(self):
                try:
                    instantiated.append(
                        DraftPipe.instantiate(value, nlp, (*path, str(i)))
                    )
                except ConfitValidationError as e:  # pragma: no cover
                    errors.append(e.raw_errors)
            if errors:
                raise ConfitValidationError(errors)
            return type(self)(instantiated)
        else:
            return self

    def _raise_draft_error(self):
        raise TypeError(
            f"This {self} component has not been instantiated "
            f"yet, likely because it was missing an `nlp` pipeline argument. You "
            f"should either:\n"
            f"- add it to a pipeline: `pipe = nlp.add_pipe(pipe)`\n"
            f"- or fill its `nlp` argument: `pipe = factory(nlp=nlp, ...)`"
        )


glob = []


def get_all_available_factories(namespaces):
    result = set()
    for ns in namespaces:
        entry_point_ns = "_".join(ns)
        eps = (
            catalogue.AVAILABLE_ENTRY_POINTS.select(group=entry_point_ns)
            if hasattr(catalogue.AVAILABLE_ENTRY_POINTS, "select")
            else catalogue.AVAILABLE_ENTRY_POINTS.get(entry_point_ns, [])
        )
        result.update([ep.name for ep in eps])
        for keys in catalogue.REGISTRY.copy().keys():
            if ns == keys[: len(ns)]:
                result.add(keys[-1])
    return sorted(result)


class FactoryRegistry(Registry):
    """
    A registry that validates the input arguments of the registered functions.
    """

    def get(self, name: str) -> Any:
        """
        Get the registered function for a given name.
        Since we want to be able to get functions registered under spacy namespace,
        and functions defined in memory but not accessible via entry points (spacy
        internal namespace), we need to check multiple registries.

        The strategy is the following:

        1. If the function exists in the edsnlp_factories namespace, return it as a
           curried function.
        2. Otherwise, check spacy namespaces and re-register it under edsnlp_factories
           namespace, then return it as a curried function.
        3. Otherwise, search in edsnlp's entry points and redo steps 1 & 2
        4. Otherwise, search in spacy's points and redo steps 1 & 2
        5. Fail

        name (str): The name.
        RETURNS (Any): The registered function.
        """

        registry_path = list(self.namespace) + [name]
        registry_spacy_path = ["spacy", "factories", name]
        registry_internal_spacy_path = ["spacy", "internal_factories", name]

        def check_and_return():
            if not catalogue.check_exists(*registry_path):
                for path in (registry_spacy_path, registry_internal_spacy_path):
                    if catalogue.check_exists(*path):
                        func = catalogue._get(path)
                        meta = spacy.Language.get_factory_meta(name)

                        self.register(
                            name,
                            func=func,
                            assigns=meta.assigns,
                            requires=meta.requires,
                            retokenizes=meta.retokenizes,
                            default_config=meta.default_config,
                        )

            if catalogue.check_exists(*registry_path):
                func = catalogue._get(registry_path)
                return lambda **kwargs: DraftPipe(func, kwargs=kwargs).maybe_nlp()

        # Steps 1 & 2
        func = check_and_return()
        if func is None and self.entry_points:
            # Update entry points in case packages lookup paths have changed
            catalogue.AVAILABLE_ENTRY_POINTS = importlib_metadata.entry_points()
            # Otherwise, step 3
            self.get_entry_point(name)
            # Then redo steps 1 & 2
            func = check_and_return()
        if func is None:
            # Otherwise, step 4
            if hasattr(spacy.registry, "_entry_point_factories"):
                spacy.registry._entry_point_factories.get_entry_point(name)
                # Then redo steps 1 & 2
                func = check_and_return()

        if func is not None:
            return func

        raise catalogue.RegistryError(
            (
                "Can't find '{name}' in registry {current_namespace}. "
                "Available names: {available_str}"
            ).format(
                name=name,
                current_namespace=" -> ".join(self.namespace),
                available_str=", ".join(
                    get_all_available_factories(
                        [
                            self.namespace,
                            ("spacy", "factories"),
                            ("spacy", "internal_factories"),
                        ]
                    )
                )
                or "none",
            )
        )

    def _get_entry_points(self) -> List[importlib_metadata.EntryPoint]:
        return (
            catalogue.AVAILABLE_ENTRY_POINTS.select(group=self.entry_point_namespace)
            if hasattr(catalogue.AVAILABLE_ENTRY_POINTS, "select")
            else catalogue.AVAILABLE_ENTRY_POINTS.get(self.entry_point_namespace, [])
        )

    def register(
        self,
        name: str,
        *,
        func: Optional[catalogue.InFunc] = None,
        default_config: Dict[str, Any] = FrozenDict(),
        assigns: Iterable[str] = FrozenList(),
        requires: Iterable[str] = FrozenList(),
        retokenizes: bool = False,
        default_score_weights: Dict[str, Optional[float]] = FrozenDict(),
        invoker: Callable = None,
        deprecated: Sequence[str] = (),
        spacy_compatible: bool = True,
    ) -> Callable[[catalogue.InFunc], catalogue.InFunc]:
        """
        This is a convenience wrapper around `confit.Registry.register`, that
        curries the function to be registered, allowing to instantiate the class
        later once `nlp` and `name` are known.

        Parameters
        ----------
        name: str
        func: Optional[catalogue.InFunc]
        default_config: Dict[str, Any]
        assigns: Iterable[str]
        requires: Iterable[str]
        retokenizes: bool
        default_score_weights: Dict[str, Optional[float]]
        invoker: Callable
        deprecated: Sequence[str]
        spacy_compatible: bool

        Returns
        -------
        Callable[[catalogue.InFunc], catalogue.InFunc]
        """
        save_params = {"@factory": name}

        def register(fn: catalogue.InFunc) -> catalogue.InFunc:
            assert (
                not spacy_compatible
                or len(accepted_arguments(fn, ["nlp", "name"])) == 2
            ), (
                "Spacy compatible factories functions must accept nlp and name as "
                "arguments. Either set register(..., spacy_compatible=False) or "
                "add nlp and name to the arguments."
            )

            meta = FactoryMeta(
                assigns=validate_attrs(assigns),
                requires=validate_attrs(requires),
                retokenizes=retokenizes,
                default_config=default_config,
            )

            def invoke(validated_fn, kwargs):
                if default_config is not None:
                    kwargs = (
                        Config(default_config)
                        .resolve(registry=self.registry)
                        .merge(kwargs)
                    )
                instantiated = (
                    invoker(validated_fn, kwargs)
                    if invoker is not None
                    else validated_fn(kwargs)
                )
                PIPE_META[instantiated] = meta
                return instantiated

            annotations = (
                fn.__annotations__
                if not isinstance(fn, type)
                else fn.__init__.__annotations__
            )
            old_annotation = None
            if "nlp" in annotations:  # annotations["nlp"] is spacy.Language:
                old_annotation = annotations["nlp"]
                annotations["nlp"] = Optional[edsnlp.core.PipelineProtocol]
            registered_fn = Registry.register(
                self,
                name=name,
                save_params=save_params,
                skip_save_params=["nlp", "name"],
                func=fn,
                invoker=invoke,
                deprecated=deprecated,
            )
            if old_annotation is not None:
                annotations["nlp"] = old_annotation

            if spacy_compatible:
                for spacy_pipe_name in (name, *deprecated):
                    spacy.Language.factory(
                        name=spacy_pipe_name,
                        default_config=default_config,
                        assigns=assigns,
                        requires=requires,
                        default_score_weights=default_score_weights,
                        retokenizes=retokenizes,
                        func=registered_fn,
                    )

            @wraps(fn)
            def curried_registered_fn(**kwargs):
                return DraftPipe(registered_fn, kwargs).maybe_nlp()

            return (
                curried_registered_fn
                if isinstance(fn, types.FunctionType)
                else registered_fn
            )

        return register(func) if func is not None else register


class registry(RegistryCollection):
    factories = FactoryRegistry(("edsnlp", "factories"), entry_points=True)
    factory = factories
    misc = Registry(("spacy", "misc"), entry_points=True)
    languages = Registry(("spacy", "languages"), entry_points=True)
    tokenizers = Registry(("spacy", "tokenizers"), entry_points=True)
    metrics = Registry(("spacy", "scorers"), entry_points=True)
    scorers = metrics
    accelerator = Registry(("edsnlp", "accelerator"), entry_points=True)
    adapters = Registry(("edsnlp", "adapters"), entry_points=True)
    readers = Registry(("edsnlp", "readers"), entry_points=True)
    writers = Registry(("edsnlp", "writers"), entry_points=True)
    core = Registry(("edsnlp", "core"), entry_points=True)
    optimizers = Registry(("edsnlp", "optimizers"), entry_points=True)
    schedules = Registry(("edsnlp", "schedules"), entry_points=True)
    loggers = Registry(("edsnlp", "loggers"), entry_points=True)


set_default_registry(registry)



import importlib
import warnings
from collections import defaultdict
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

import regex
import torch
import torch.optim
from confit import Config, validate_arguments
from confit.utils.collections import split_path
from typing_extensions import Literal

import edsnlp
from edsnlp.core import PipelineProtocol
from edsnlp.utils.collections import get_deep_attr, set_deep_attr

optim_mapping = {
    "adam": torch.optim.Adam,
    "adamw": torch.optim.AdamW,
    "sgd": torch.optim.SGD,
    "adadelta": torch.optim.Adadelta,
    "adagrad": torch.optim.Adagrad,
    "adamax": torch.optim.Adamax,
    "rmsprop": torch.optim.RMSprop,
}


def get_optimizer(optim):
    if isinstance(optim, str):
        optim_lower = optim.lower()
        if optim_lower in optim_mapping:
            return optim_mapping[optim_lower]
        else:
            try:
                # Attempt to get the optimizer from torch.optim
                return getattr(torch.optim, optim)
            except AttributeError:
                # If not found in torch.optim, try to import it dynamically
                module_name, class_name = optim.rsplit(".", 1)
                module = importlib.import_module(module_name)
                return getattr(module, class_name)
    return optim


@validate_arguments
class Schedule:
    paths: List[Tuple[Union[str, int]]]
    start_value: Any

    def __init__(self, path, start_value: Any):
        self.paths: Optional[List[Tuple[Union[str, int]]]] = (
            None
            if path is None
            else [(path if isinstance(path, list) else split_path(path))]
        )
        self.start_value = start_value

    def step(self, group, closure=None):
        raise NotImplementedError

    def reset(self, group):
        raise NotImplementedError()

    def state_dict(self):
        raise NotImplementedError()

    def load_state_dict(self, state):
        raise NotImplementedError()


@edsnlp.registry.schedules.register("linear")
class LinearSchedule(Schedule):
    def __init__(
        self,
        total_steps: Optional[int] = None,
        max_value: Optional[Any] = None,
        start_value: float = 0.0,
        path: Optional[Union[str, int, List[Union[str, int]]]] = None,
        warmup_rate: float = 0.0,
        end_value: float = 0.0,
    ):
        """
        Linear schedule for a parameter group. The schedule will linearly increase
        the value from `start_value` to `max_value` in the first `warmup_rate` of the
        `total_steps` and then linearly decrease it to `end_value`.

        Parameters
        ----------
        total_steps: Optional[int]
            The total number of steps, usually used to calculate ratios.
        max_value: Optional[Any]
            The maximum value to reach.
        start_value: float
            The initial value.
        path: str
            The path to the attribute to set.
        warmup_rate: float
            The rate of the warmup.
        end_value: float
            The final value to reach after the decay phase. Defaults to 0.0.
        """
        super().__init__(path, start_value)
        self.max_value = max_value
        self.warmup_rate = warmup_rate
        self.total_steps = total_steps
        self.end_value = end_value
        self.idx = 0

    def reset(self, group):
        self.idx = -1
        self.step(group)

    def state_dict(self):
        return {
            "idx": self.idx,
        }

    def load_state_dict(self, state):
        self.idx = state["idx"]

    def step(self, group, closure=None):
        self.idx += 1
        if self.max_value is None:
            if isinstance(get_deep_attr(group, self.paths[0]), (int, float)):
                self.max_value = get_deep_attr(group, self.paths[0])
            else:
                raise Exception(
                    "The max_value parameter of the linear schedule "
                    "must be set to a valid number."
                )
        warmup_steps = self.total_steps * self.warmup_rate
        if self.idx < warmup_steps:
            progress = self.idx / warmup_steps
            value = self.start_value + (self.max_value - self.start_value) * progress
        else:
            progress = min(
                1.0, (self.idx - warmup_steps) / (self.total_steps - warmup_steps)
            )
            value = self.max_value + (self.end_value - self.max_value) * progress
        for path in self.paths:
            set_deep_attr(group, path, value)

    def __repr__(self):
        format_string = type(self).__name__ + "(\n"
        format_string += f"    start_value: {self.start_value}\n"
        format_string += f"    max_value: {self.max_value}\n"
        format_string += f"    end_value: {self.end_value}\n"
        format_string += f"    warmup_rate: {self.warmup_rate}\n"
        format_string += f"    paths: {self.paths}\n"
        format_string += f"    total_steps: {self.total_steps}\n"
        format_string += ")"
        return format_string


@edsnlp.registry.core.register("optimizer")
class ScheduledOptimizer(torch.optim.Optimizer):
    def __init__(
        self,
        optim: Union[torch.optim.Optimizer, Type[torch.optim.Optimizer], str],
        module: Optional[Union[PipelineProtocol, torch.nn.Module]] = None,
        total_steps: Optional[int] = None,
        groups: Optional[
            Union[List[Dict], Dict[str, Union[Dict, Literal[False]]]]
        ] = None,
        init_schedules: bool = True,
        **kwargs,
    ):
        """
        Wrapper optimizer that supports schedules for the parameters and easy parameter
        selection using the key of the `groups` dictionary as regex patterns to match
        the parameter names.

        Schedules are defined directly in the groups, in place of the scheduled value.

        Examples
        --------
        ```{ .python .no-check }
        optim = ScheduledOptimizer(
            cls="adamw",
            module=model,
            groups=[
                # Exclude all parameters matching 'bias' from optimization.
                {
                    "selector": "bias",
                    "exclude": True,
                },
                # Parameters of the NER module's embedding receive this learning rate
                # schedule. If a parameter matches both 'transformer' and 'ner',
                # the first group settings take precedence due to the order.
                {
                    "selector": "^ner[.]embedding"
                    "lr": {
                        "@schedules": "linear",
                        "start_value": 0.0,
                        "max_value": 5e-4,
                        "warmup_rate": 0.2,
                    },
                },
                # Parameters starting with 'ner' receive this learning rate schedule,
                # unless a 'lr' value has already been set by an earlier selector.
                {
                    "selector": "^ner"
                    "lr": {
                        "@schedules": "linear",
                        "start_value": 0.0,
                        "max_value": 1e-4,
                        "warmup_rate": 0.2,
                    },
                },
                # Apply a weight_decay of 0.01 to all parameters not excluded.
                # This setting doesn't conflict with others and applies to all.
                {
                    "selector": "",
                    "weight_decay": 0.01,
                },
            ],
            total_steps=1000,
        )
        ```

        Parameters
        ----------
        optim : Union[str, Type[torch.optim.Optimizer], torch.optim.Optimizer]
            The optimizer to use. If a string (like "adamw") or a type to instantiate,
            the `module` and `groups` must be provided.
        module : Optional[Union[PipelineProtocol, torch.nn.Module]]
            The module to optimize. Usually the `nlp` pipeline object.
        total_steps : Optional[int]
            The total number of steps, used for schedules.
        groups : Optional[List[Group]]
            The groups to optimize. Each group is a dictionary containing:

            - a regex `selector` key to match the parameter of that group by their names
              (as listed by `nlp.named_parameters()`)
            - and several other keys that define the optimizer parameters for that
              group, such as `lr`, `weight_decay` etc. The value for these keys can
              be a `Schedule` instance or a simple value
            - an `exclude` key that can be set to True to exclude parameters

            The matching is performed by running `regex.search(selector, name)` so you
            do not have to match the full name. Note that the order of the groups
            matters. If a parameter name matches multiple selectors, the
            configurations of these selectors are combined in reverse order (from the
            last matched selector to the first), allowing later selectors to complete
            options from earlier ones. If a selector contains `exclude=True`, any
            parameter matching it is excluded from optimization.
        """
        should_instantiate_optim = isinstance(optim, str) or isinstance(optim, type)
        if should_instantiate_optim and (groups is None or module is None):
            raise ValueError(
                "If the optimizer is a string or a type, the module and groups must "
                "be provided."
            )
        elif not should_instantiate_optim and (
            groups is not None or module is not None
        ):
            raise ValueError(
                "If the optimizer is already instantiated, the module and groups must "
                "not be provided."
            )

        if should_instantiate_optim:
            named_parameters = list(module.named_parameters())
            groups = Config.resolve(groups, registry=edsnlp.registry)

            # New groups format
            if isinstance(groups, list):
                groups = [dict(g) for g in groups]
                groups = {
                    g.pop("selector"): g if not g.get("exclude") else False
                    for g in groups
                }

            groups = {
                sel: dict(group) if group else False for sel, group in groups.items()
            }
            param_to_groups = {}
            for name, param in named_parameters:
                param_to_groups[param] = tuple(
                    dict.fromkeys(
                        sel for sel, group in groups.items() if regex.search(sel, name)
                    )
                )
            groups_to_params = defaultdict(lambda: [])
            empty_selectors = {sel for sel in groups}
            for params, group in param_to_groups.items():
                groups_to_params[group].append(params)
                for sel in group:
                    empty_selectors.discard(sel)

            if empty_selectors:
                warnings.warn(
                    f"Selectors {list(empty_selectors)} did not match any parameters."
                )
                warnings.warn(
                    "For reference, here are the parameters of the module:\n"
                    + "\n".join("- " + name for name, _ in named_parameters)
                )

            cliques = []
            for selectors, params in groups_to_params.items():
                group = {}
                group_sources = {}
                for sel in reversed(selectors):
                    if groups[sel] is False:
                        break
                    group.update(groups[sel])
                    group_sources.update({k: sel for k in groups[sel]})
                else:
                    # if no group=False (break) was encountered
                    if group and "lr" in group and params:
                        sources = [
                            sel for sel in selectors if sel in group_sources.values()
                        ]
                        group["selectors"] = sources
                        group["params"] = params
                        cliques.append(group)
            cliques = reversed(
                [{k: v for k, v in group.items() if v is not None} for group in cliques]
            )

            optim = get_optimizer(optim)
            optim = optim(cliques, **kwargs)

        self.optim = optim
        self.schedules = self.extract_schedules(optim.param_groups)
        for schedule in self.schedules:
            if schedule.total_steps is None:
                assert total_steps is not None, (
                    "total_steps must be provided to the optimizer or the schedule"
                )
                schedule.total_steps = total_steps
            if init_schedules:
                schedule.step(optim.param_groups)

    @classmethod
    def extract_schedules(cls, param_groups):
        schedules = defaultdict(set)

        def rec(node, path):
            if len(path) == 2 and path[1] in ("schedules", "params"):
                return
            if isinstance(node, dict):
                items = node.items()
            elif isinstance(node, (list, tuple)):
                items = enumerate(node)
            else:
                if isinstance(node, Schedule):
                    schedules[node].add(path)
                return
            for key, value in items:
                rec(value, (*path, key))

        # For backward compatibility when schedules were defined
        # under the "schedules" key in groups.
        for i, group in enumerate(param_groups):
            if "schedules" in group:
                grp_schedules = group["schedules"]
                grp_schedules = (
                    grp_schedules
                    if isinstance(grp_schedules, list)
                    else [grp_schedules]
                )
                for schedule in grp_schedules:
                    if schedule.paths is None:
                        schedule.paths = [("lr",)]
                    schedule.paths = [(i, *path) for path in schedule.paths]
                    schedules[schedule].update(schedule.paths)

        rec(param_groups, ())

        for schedule, paths in schedules.items():
            paths = sorted(paths)
            if schedule.paths is None:
                schedule.paths = paths
            if schedule.paths != paths:
                raise ValueError(f"Schedule path mismatch: {schedule.paths} != {paths}")

        return list(schedules.keys())

    def zero_grad(self):
        return self.optim.zero_grad()

    @property
    def param_groups(self):
        return self.optim.param_groups

    @param_groups.setter
    def param_groups(self, value):
        self.optim.param_groups = value

    @property
    def state(self):
        return self.optim.state

    @state.setter
    def state(self, value):
        self.optim.state = value

    def state_dict(self):
        state = {
            "optim": self.optim.state_dict(),
            "schedules": [schedule.state_dict() for schedule in self.schedules],
        }
        return state

    def load_state_dict(self, state):
        self.optim.load_state_dict(state["optim"])
        for schedule_state, schedule in zip(state["schedules"], self.schedules):
            schedule.load_state_dict(schedule_state)

    def step(self, closure=None):
        self.optim.step(closure=closure)
        self.step_schedules()

    def step_schedules(self):
        for schedule in self.schedules:
            schedule.step(self.param_groups)

    def initialize(self):
        self.reset()
        self.optim.step()

    def reset(self):
        self.optim.zero_grad()
        for group in self.optim.param_groups:
            for param in group["params"]:
                if param.requires_grad:
                    param.grad = torch.zeros_like(param)
        for schedule in self.schedules:
            schedule.reset(self.param_groups)

    def __repr__(self):
        format_string = type(self).__name__ + f"[{type(self.optim).__qualname__}] ("
        ind = "    "
        for i, group in enumerate(self.param_groups):
            format_string += "\n"
            format_string += f"Parameter Group {i}\n"
            keys = [
                "selector",
                "params",
                "lr",
                *sorted(set(group.keys()) - {"selector", "params", "lr"}),
            ]
            for key in keys:
                if key in group:
                    format_string += ind + f"{key}: "
                    if key == "params":
                        num_tensors = len(group["params"])
                        num_params = sum(p.numel() for p in group["params"])
                        format_string += (
                            f"{num_params} parameters ({num_tensors} tensors)\n"
                        )
                    elif isinstance(group[key], list):
                        format_string += "[\n"
                        for i, item in enumerate(group[key]):
                            sub_str = str(item)
                            for line in sub_str.split("\n"):
                                format_string += ind * 2 + line + "\n"
                        format_string += ind + "]\n"
                    else:
                        sub_str = str(group[key])
                        for i, line in enumerate(sub_str.split("\n")):
                            format_string += (ind if i > 0 else "") + line + "\n"

        format_string += ")"
        return format_string



from __future__ import annotations

import logging
import os
import pickle
import warnings
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import foldedtensor as ft
import torch
import torch.nn.functional as F
from spacy.tokens import Doc, Span
from typing_extensions import NotRequired, TypedDict

from edsnlp.core import PipelineProtocol
from edsnlp.core.torch_component import BatchInput, BatchOutput, TorchComponent
from edsnlp.pipes.base import BaseSpanAttributeClassifierComponent
from edsnlp.pipes.trainable.embeddings.typing import (
    SpanEmbeddingComponent,
)
from edsnlp.utils.bindings import (
    BINDING_GETTERS,
    BINDING_SETTERS,
    Attributes,
    AttributesArg,
)
from edsnlp.utils.span_getters import (
    ContextWindow,
    SpanFilter,
    SpanGetterArg,
    get_spans,
)

logger = logging.getLogger(__name__)

SpanClassifierBatchInput = TypedDict(
    "SpanClassifierBatchInput",
    {
        "embedding": BatchInput,
        "targets": NotRequired[torch.Tensor],
    },
)
"""
embeds: torch.FloatTensor
    Token embeddings to predict the tags from
mask: torch.BoolTensor
    Mask of the sequences
spans: torch.Tensor
    2d tensor of n_spans * (doc_idx, ner_label_idx, begin, end)
targets: NotRequired[List[torch.Tensor]]
    list of 2d tensor of n_spans * n_combinations (1 hot)
"""

SpanClassifierBatchOutput = TypedDict(
    "SpanClassifierBatchOutput",
    {
        "loss": Optional[torch.Tensor],
        "labels": Optional[List[torch.Tensor]],
    },
)
"""
loss: Optional[torch.Tensor]
    The loss of the model
labels: Optional[List[torch.Tensor]]
    The predicted labels
"""


class TrainableSpanClassifier(
    TorchComponent[BatchOutput, SpanClassifierBatchInput],
    BaseSpanAttributeClassifierComponent,
):
    """
    The `eds.span_classifier` component is a trainable attribute predictor.
    In this context, the span classification task consists in assigning values (boolean,
    strings or any object) to attributes/extensions of spans such as:

    - `span._.negation`,
    - `span._.date.mode`
    - `span._.cui`

    In the rest of this page, we will refer to a pair of (attribute, value) as a
    "binding". For instance, the binding `("_.negation", True)` means that the
    attribute `negation` of the span is (or should be, when predicted) set to `True`.

    Architecture
    ------------
    The model performs span classification by:

    1. Calling a word pooling embedding such as `eds.span_pooler` to compute a single
    embedding for each span
    2. Computing logits for each possible binding using a linear layer
    3. Splitting these bindings into groups of exclusive values such as

        - `event=start` and `event=stop`
        - `negated=False` and `negated=True`

        Note that the above groups are not exclusive, but the values within each group
        are.

    4. Applying the best scoring binding in each group to each span

    Examples
    --------
    To create a span classifier component, you can use the following code:

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.span_classifier(
            # To embed the spans, we will use a span pooler
            embedding=eds.span_pooler(
                pooling_mode="mean",  # mean pooling
                # that will use a transformer to embed the doc words
                embedding=eds.transformer(
                    model="prajjwal1/bert-tiny",
                    window=128,
                    stride=96,
                ),
            ),
            span_getter=["ents", "sc"],
            # For every span embedded by the span pooler
            # (doc.ents and doc.spans["sc"]), we will predict both
            # span._.negation and span._.event_type
            attributes=["_.negation", "_.event_type"],
        ),
        name="span_classifier",
    )
    ```

    To infer the values of the attributes, you can use the pipeline `post_init` method:

    ```{ .python .no-check }
    nlp.post_init(gold_data)
    ```

    To train the model, refer to the [Training](/tutorials/training-span-classifier)
    tutorial.

    You can inspect the bindings that will be used for training and prediction
    ```{ .python .no-check }
    print(nlp.pipes.attr.bindings)
    # list of (attr name, span labels or True if all, values)
    # Out: [
    #   ('_.negation', True, [True, False]),
    #   ('_.event_type', True, ['start', 'stop'])
    # ]
    ```

    You can also change these values and update the bindings by calling the
    `update_bindings` method. Don't forget to retrain the model if new values are
    added !

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : str
        Name of the component
    embedding : SpanEmbeddingComponent
        The word embedding component
    label_weights: Dict[str, Dict[Any, float]]
        The weight of each label for each attribute. The keys are the attribute names
        and the values are dictionaries with the labels as keys and the weights as
        values. For instance, `{"_.negation": {True: 1, False: 2}}` will give a weight
        of 1 to the `True` value of the `negation` attribute and 2 to the `False` value.
    span_getter : SpanGetterArg
        How to extract the candidate spans and the attributes to predict or train on.
    context_getter : Optional[Union[Callable, SpanGetterArg]]
        What context to use when computing the span embeddings (defaults to the whole
        document). This can be:

        - a `SpanGetterArg` to retrieve contexts from a whole document. For example
          `{"section": "conclusion"}` to only use the conclusion as context (you
          must ensure that all spans produced by the `span_getter` argument do fall
          in the conclusion in this case)
        - a callable, that gets a span and should return a context for this span.
          For instance, `lambda span: span.sent` to use the sentence as context.
    attributes : AttributesArg
        The attributes to predict or train on. If a dict is given, keys are the
        attributes and values are the labels for which the attr is allowed, or True
        if the attr is allowed for all labels.
    keep_none : bool
        If False, skip spans for which a attr returns None. If True (default), the
        None values will be learned and predicted, just as any other value.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        name: str = "span_classifier",
        *,
        embedding: SpanEmbeddingComponent,
        attributes: AttributesArg = None,
        qualifiers: AttributesArg = None,
        label_weights: Dict[str, Dict[Any, float]] = None,
        span_getter: SpanGetterArg = None,
        context_getter: Optional[Union[ContextWindow, SpanGetterArg]] = None,
        values: Optional[Dict[str, List[Any]]] = None,
        keep_none: bool = False,
    ):
        attributes: Attributes
        if attributes is None and qualifiers is None:
            raise TypeError(
                "The `attributes` parameter is required. Please provide a dict of "
                "attributes to predict or train on."
            )

        if qualifiers is not None:
            warnings.warn(
                "The `qualifiers` parameter is deprecated. Use `attributes` instead."
            )
            assert attributes is None
            attributes = qualifiers
        sub_span_getter = getattr(embedding, "span_getter", None)
        if (
            sub_span_getter is not None and span_getter is None
        ):  # pragma: no cover  # noqa: E501
            span_getter = sub_span_getter
        span_getter = span_getter or {"ents": True}
        sub_context_getter = getattr(embedding, "context_getter", None)
        if (
            sub_context_getter is not None and context_getter is None
        ):  # pragma: no cover
            context_getter = sub_context_getter

        self.values = values
        self.keep_none = keep_none

        attributes = {
            k if k.startswith("_.") else f"_.{k}": v for k, v in attributes.items()
        }
        self.label_weights_bindings = (
            {
                (attr if attr.startswith("_.") else f"_.{attr}", value): weight
                for attr, values in label_weights.items()
                for value, weight in values.items()
            }
            if label_weights
            else dict()
        )

        unknown = set([attr for attr, _ in self.label_weights_bindings]) - set(
            attributes.keys()
        )
        if unknown:
            warnings.warn(
                f"Attributes ({unknown}) are present in label_weights "
                f"but not in attributes: Those weights will be ignored"
            )

        self.bindings: List[Tuple[str, List[str], List[Any]]] = [
            (k, v, []) for k, v in attributes.items()
        ]

        super().__init__(nlp, name, span_getter=span_getter)
        self.embedding = embedding
        self.context_getter = context_getter
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            self.classifier = torch.nn.Linear(embedding.output_size, 0)

    @property
    def attributes(self) -> Attributes:
        return {qlf: labels for qlf, labels, _ in self.bindings}

    @attributes.setter
    def attributes(self, value: Attributes):
        bindings = []
        for qlf, labels in value.items():
            groups = [group for group in self.bindings if group[0] == qlf]
            if len(groups) > 1:
                raise ValueError(
                    f"Attribute {qlf} has different label filters: "
                    f"{[g[0] for g in groups]}. Please use the `update_bindings` "
                    f"method to update the labels."
                )
            if groups:
                bindings.append((qlf, labels, groups[0][2]))
        self.update_bindings(bindings)

    def to_disk(self, path, *, exclude=set()):
        repr_id = object.__repr__(self)
        if repr_id in exclude:
            return
        # This will receive the directory path + /my_component
        # We save the bindings as a pickle file since values can be arbitrary objects
        os.makedirs(path, exist_ok=True)
        data_path = path / "bindings.pkl"
        with open(data_path, "wb") as f:
            pickle.dump(
                {
                    "bindings": self.bindings,
                },
                f,
            )
        return super().to_disk(path, exclude=exclude)

    def from_disk(self, path, exclude=tuple()):
        repr_id = object.__repr__(self)
        if repr_id in exclude:
            return
        # This will receive the directory path + /my_component
        data_path = path / "bindings.pkl"
        with open(data_path, "rb") as f:
            data = pickle.load(f)
        self.update_bindings(data["bindings"])
        self.set_extensions()
        super().from_disk(path, exclude=exclude)

    @property
    def bindings_to_idx(self) -> List[Tuple[str, List[str], Dict[Any, int]]]:
        if getattr(self, "_bindings_to_idx", None) is not None:
            return self._bindings_to_idx
        self._bindings_to_idx = [  # noqa
            (attr, labels, {value: idx for idx, value in enumerate(values)})
            for attr, labels, values in self.bindings
        ]
        return self._bindings_to_idx

    @property
    def bindings_indexers(self) -> List[Union[Sequence[int], slice]]:
        return self._bindings_indexers

    def set_extensions(self):
        super().set_extensions()
        for group in self.bindings:
            qlf = group[0]
            if qlf.startswith("_."):
                qlf = qlf[2:]
            if not Span.has_extension(qlf):
                Span.set_extension(qlf, default=None)

    def post_init(self, gold_data: Iterable[Doc], exclude: Set[str]):
        super().post_init(gold_data, exclude=exclude)

        bindings = [
            (qlf, labels, dict.fromkeys(vals)) for qlf, labels, vals in self.bindings
        ]
        for doc in gold_data:
            spans = list(get_spans(doc, self.span_getter))
            for span in spans:
                for attr, labels, values in bindings:
                    if labels is True or span.label_ in labels:
                        value = BINDING_GETTERS[attr](span)
                        if value is not None or self.keep_none:
                            values[value] = None

        bindings = [
            (attr, labels, sorted(values, key=str)) for attr, labels, values in bindings
        ]

        for attr, labels, values in bindings:
            if len(values) < 2:
                warnings.warn(
                    f"Attribute {attr} for labels {labels} should have at "
                    f"least 2 values but found {len(values)}: {values}."
                )

        self.update_bindings(bindings)

    def update_bindings(self, bindings: List[Tuple[str, SpanFilter, List[Any]]]):
        keep_bindings = [
            (labels is True or len(labels) > 0) and len(values) > 0
            for k, labels, values in bindings
        ]
        if not all(keep_bindings):
            logger.warning(
                "Some attributes have no labels or values and have been removed:"
                + "".join(
                    "\n- " + str(b[0]) + " for labels " + str(b[1])
                    for b, keep in zip(bindings, keep_bindings)
                    if not keep
                )
            )
        bindings = [b for b, keep in zip(bindings, keep_bindings) if keep]
        new_bindings = list(
            dict.fromkeys((k, v) for k, _, group in bindings for v in group)
        )
        old_bindings = list(
            dict.fromkeys((k, v) for k, _, group in self.bindings for v in group)
        )

        new_bindings_to_idx = {binding: idx for idx, binding in enumerate(new_bindings)}
        old_bindings_to_idx = {binding: idx for idx, binding in enumerate(old_bindings)}

        common = [b for b in new_bindings if b in old_bindings]
        old_index = [old_bindings_to_idx[label] for label in common]
        new_index = [new_bindings_to_idx[label] for label in common]
        new_linear = torch.nn.Linear(self.classifier.in_features, len(new_bindings))
        new_linear.weight.data[new_index] = self.classifier.weight.data[old_index]
        self.classifier.weight.data = new_linear.weight.data
        self.classifier.out_features = new_bindings
        missing_bindings = set(new_bindings) - set(old_bindings)
        if missing_bindings and len(old_bindings) > 0:
            warnings.warn(
                f"Added {len(missing_bindings)} new bindings. Consider retraining "
                f"the model to learn these new bindings."
            )

        if hasattr(self.classifier, "bias"):
            new_linear.bias.data[new_index] = self.classifier.bias.data[old_index]
            self.classifier.bias.data = new_linear.bias.data

        def simplify_indexer(indexer):
            return (
                slice(indexer[0], indexer[-1] + 1)
                if len(indexer) and list(range(indexer[0], indexer[-1] + 1)) == indexer
                else indexer
            )

        self.bindings = bindings
        self._bindings_indexers = [
            simplify_indexer([new_bindings_to_idx[(qlf, value)] for value in values])
            for qlf, labels, values in bindings
        ]
        self.label_weights = [
            self.label_weights_bindings.get((qlf, value), 1)
            for qlf, labels, values in bindings
            for value in values
        ]
        self._bindings_to_idx = None

    def preprocess(self, doc: Doc, **kwargs) -> Dict[str, Any]:
        spans = list(get_spans(doc, self.span_getter))
        if self.context_getter is None or not callable(self.context_getter):
            contexts = list(get_spans(doc, self.context_getter))
            pre_aligned = False
        else:
            contexts = [self.context_getter(span) for span in spans]
            pre_aligned = True
        return {
            "embedding": self.embedding.preprocess(
                doc,
                spans=spans,
                contexts=contexts,
                pre_aligned=pre_aligned,
                **kwargs,
            ),
            "$spans": spans,
        }

    def preprocess_supervised(self, doc: Doc) -> Dict[str, Any]:
        preps = self.preprocess(doc)
        return {
            **preps,
            "targets": [
                [
                    values_to_idx.get(BINDING_GETTERS[qlf](span), -100)
                    if labels is True or span.label_ in labels
                    else -100
                    for qlf, labels, values_to_idx in self.bindings_to_idx
                ]
                for span in preps["$spans"]
            ],
        }

    def collate(self, batch: Dict[str, Sequence[Any]]) -> SpanClassifierBatchInput:
        collated: SpanClassifierBatchInput = {
            "embedding": self.embedding.collate(batch["embedding"]),
        }
        if "targets" in batch:
            targets = ft.as_folded_tensor(
                batch["targets"],
                dtype=torch.long,
                full_names=("sample", "span", "group"),
                data_dims=("span", "group"),
            ).as_tensor()
            collated["targets"] = targets.view(len(targets), len(self.bindings))
        return collated

    # noinspection SpellCheckingInspection
    def forward(self, batch: SpanClassifierBatchInput) -> BatchOutput:
        """
        Apply the span classifier module to the document embeddings and given spans to:
        - compute the loss
        - and/or predict the labels of spans

        Parameters
        ----------
        batch: SpanClassifierBatchInput
            The input batch

        Returns
        -------
        BatchOutput
        """
        embedding = self.embedding(batch["embedding"])
        span_embeds = embedding["embeddings"]

        binding_scores = self.classifier(span_embeds)

        if "targets" in batch:
            losses = []
            pred = None
        else:
            pred = []
            losses = None

        # For each group, for instance:
        # - `event=start` and `event=stop`
        # - `negated=False` and `negated=True`
        for group_idx, bindings_indexer in enumerate(self.bindings_indexers):
            if "targets" in batch:
                losses.append(
                    F.cross_entropy(
                        binding_scores[:, bindings_indexer],
                        batch["targets"][:, group_idx],
                        reduction="sum",
                        weight=torch.tensor(self.label_weights, dtype=torch.float)[
                            bindings_indexer
                        ].to(binding_scores.device),
                    )
                )
                assert not torch.isnan(losses[-1]).any(), "NaN loss"
            else:
                pred.append(binding_scores[:, bindings_indexer].argmax(dim=1))

        return {
            "loss": sum(losses) if losses is not None else None,
            "labels": pred,
        }

    def postprocess(
        self,
        docs: Sequence[Doc],
        results: SpanClassifierBatchOutput,
        inputs: List[Dict[str, Any]],
    ) -> Sequence[Doc]:
        # Preprocessed docs should still be in the cache
        spans = [span for sample in inputs for span in sample["$spans"]]
        all_labels = results["labels"]
        # For each prediction group (exclusive bindings)...
        for val_indices, (qlf, labels, values) in zip(all_labels, self.bindings):
            # For each span...
            for span, idx in zip(spans, val_indices.tolist()):
                # If the span is not filtered out...
                if labels is True or span.label_ in labels:
                    # ...assign the predicted value to the span
                    BINDING_SETTERS[qlf](span, values[idx])
        return docs


# For backward compatibility
TrainableSpanQualifier = TrainableSpanClassifier



# ruff:noqa:E501

number_terms = {
    "0.125": ["⅛"],
    "0.16666666": ["⅙"],
    "0.2": ["⅕"],
    "0.25": ["¼"],
    "0.3333333": ["⅓"],
    "0.5": ["½"],
    "2.5": ["21/2"],
    "1": ["un", "une"],
    "2": ["deux"],
    "3": ["trois"],
    "4": ["quatre"],
    "5": ["cinq"],
    "6": ["six"],
    "7": ["sept"],
    "8": ["huit"],
    "9": ["neuf"],
    "10": ["dix"],
    "11": ["onze"],
    "12": ["douze"],
    "13": ["treize"],
    "14": ["quatorze"],
    "15": ["quinze"],
    "16": ["seize"],
    "17": ["dix-sept", "dix sept"],
    "18": ["dix-huit", "dix huit"],
    "19": ["dix-neuf", "dix neuf"],
    "20": ["vingt", "vingts"],
    "30": ["trente"],
    "40": ["quarante"],
    "50": ["cinquante"],
    "60": ["soixante"],
    "70": ["soixante dix", "soixante-dix"],
    "80": ["quatre vingt", "quatre-vingt", "quatre vingts", "quatre-vingts"],
    "90": ["quatre vingt dix", "quatre-vingt-dix"],
    "100": ["cent"],
    "500": ["cinq cent", "cinq-cent"],
    "1000": ["mille", "milles"],
}


number_regex = r"""(?x)
# no digit or floating point number prefix before
(?<![0-9][.,]?)
# integer part like 123 or 1 234
(?:
    0
    |[1-9][0-9]*(?:\ \d{3})*
)
(?:
    # floating point surounded by spaces
    \ +[,.]\ +\d+
    # floating point w/o space
    | [,.]\d+
    # fractions
    | (?:\ /\ |/)[1-9][0-9]*(?:\ \d{3})*
)?"""


common_quantities = {
    "weight": {
        "unit": "kg",
        "unitless_patterns": [
            {
                "terms": ["poids", "poid", "pese", "pesant", "pesait", "pesent"],
                "ranges": [
                    {"min": 0, "max": 200, "unit": "kg"},
                    {"min": 200, "unit": "g"},
                ],
            }
        ],
    },
    "size": {
        "unit": "m",
        "unitless_patterns": [
            {
                "terms": [
                    "mesure",
                    "taille",
                    "mesurant",
                    "mesurent",
                    "mesurait",
                    "mesuree",
                    "hauteur",
                    "largeur",
                    "longueur",
                ],
                "ranges": [
                    {"min": 0, "max": 3, "unit": "m"},
                    {"min": 3, "unit": "cm"},
                ],
            }
        ],
    },
    "bmi": {
        "unit": "kg_per_m2",
        "unitless_patterns": [
            {"terms": ["imc", "bmi"], "ranges": [{"unit": "kg_per_m2"}]}
        ],
    },
    "volume": {"unit": "m3", "unitless_patterns": []},
}

unit_divisors = ["/", "par"]

stopwords = ["par", "sur", "de", "a", ",", "et", "-", "à"]

# Should we only make accented patterns and expect the user to use
# `eds.normalizer` component first ?
range_patterns = [
    ("De", "à"),
    ("De", "a"),
    ("de", "à"),
    ("de", "a"),
    ("Entre", "et"),
    ("entre", "et"),
    (None, "a"),
    (None, "à"),
    (None, "-"),
]

# fmt: off
units_config = {
    # Mass
    "kg":  {"dim": "mass", "degree": 1, "scale": 1e3, "terms": ["kgramme", "kilogramme", "kilo-gramme", "kgrammes", "kilogrammes", "kilo-grammes", "kgr", "kilogr", "kilo-gr", "kg", "kilog", "kilo-g"], "followed_by": "g"},
    "hg":  {"dim": "mass", "degree": 1, "scale": 1e2, "terms": ["hgramme", "hectogramme", "hecto-gramme", "hgrammes", "hectogrammes", "hecto-grammes", "hgr", "hectogr", "hecto-gr", "hg", "hectog", "hecto-g"], "followed_by": None},
    "dag": {"dim": "mass", "degree": 1, "scale": 1e1, "terms": ["dagramme", "decagramme", "deca-gramme", "dagrammes", "decagrammes", "deca-grammes", "dagr", "decagr", "deca-gr", "dag", "decag", "deca-g"], "followed_by": None},
    "g":   {"dim": "mass", "degree": 1, "scale": 1e0, "terms": ["gramme", "grammes", "gr", "g"], "followed_by": None},
    "dg":  {"dim": "mass", "degree": 1, "scale": 1e-1, "terms": ["dgramme", "decigramme", "deci-gramme", "dgrammes", "decigrammes", "deci-grammes", "dgr", "decigr", "deci-gr", "dg", "decig", "deci-g"], "followed_by": None},
    "cg":  {"dim": "mass", "degree": 1, "scale": 1e-2, "terms": ["cgramme", "centigramme", "centi-gramme", "cgrammes", "centigrammes", "centi-grammes", "cgr", "centigr", "centi-gr", "cg", "centig", "centi-g"], "followed_by": None},
    "mg":  {"dim": "mass", "degree": 1, "scale": 1e-3, "terms": ["mgramme", "milligramme", "milli-gramme", "miligramme", "mili-gramme", "mgrammes", "milligrammes", "milli-grammes", "miligrammes", "mili-grammes", "mgr", "milligr", "milli-gr", "miligr", "mili-gr", "mg", "millig", "milli-g", "milig", "mili-g"], "followed_by": None},
    "μg":  {"dim": "mass", "degree": 1, "scale": 1e-6, "terms": ["μgramme", "ugramme", "µgramme", "microgramme", "micro-gramme", "μgrammes", "ugrammes", "µgrammes", "microgrammes", "micro-grammes", "μgr", "ugr", "µgr", "microgr", "micro-gr", "μg", "ug", "µg", "microg", "micro-g"], "followed_by": None},
    "ng":  {"dim": "mass", "degree": 1, "scale": 1e-9, "terms": ["ngramme", "nanogramme", "nano-gramme", "ngrammes", "nanogrammes", "nano-grammes", "ngr", "nanogr", "nano-gr", "ng", "nanog", "nano-g"], "followed_by": None},
    "pg":  {"dim": "mass", "degree": 1, "scale": 1e-12, "terms": ["pgramme", "picogramme", "pico-gramme", "pgrammes", "picogrammes", "pico-grammes", "pgr", "picogr", "pico-gr", "pg", "picog", "pico-g"], "followed_by": None},
    "fg":  {"dim": "mass", "degree": 1, "scale": 1e-15, "terms": ["fgramme", "femtogramme", "femto-gramme", "fgrammes", "femtogrammes", "femto-grammes", "fgr", "femtogr", "femto-gr", "fg", "femtog", "femto-g"], "followed_by": None},

    # Mass^-1
    "per_kg": {"dim": "mass", "degree": -1, "scale": 1e-3, "terms": ["kg-1", "kilog-1", "kilo-g-1", "kgr-1", "kilogr-1", "kilo-gr-1", "kgr⁻¹", "kilogr⁻¹", "kilo-gr⁻¹", "kg⁻¹", "kilog⁻¹", "kilo-g⁻¹"], "followed_by": None},
    "per_hg": {"dim": "mass", "degree": -1, "scale": 1e-2, "terms": ["hg-1", "hectog-1", "hecto-g-1", "hgr-1", "hectogr-1", "hecto-gr-1", "hgr⁻¹", "hectogr⁻¹", "hecto-gr⁻¹", "hg⁻¹", "hectog⁻¹", "hecto-g⁻¹"], "followed_by": None},
    "per_dag": {"dim": "mass", "degree": -1, "scale": 1e-1, "terms": ["dag-1", "decag-1", "deca-g-1", "dagr-1", "decagr-1", "deca-gr-1", "dagr⁻¹", "decagr⁻¹", "deca-gr⁻¹", "dag⁻¹", "decag⁻¹", "deca-g⁻¹"], "followed_by": None},
    "per_g": {"dim": "mass", "degree": -1, "scale": 1e0, "terms": ["g-1", "gr-1", "gr⁻¹", "g⁻¹"], "followed_by": None},
    "per_dg": {"dim": "mass", "degree": -1, "scale": 1e1, "terms": ["dg-1", "decig-1", "deci-g-1", "dgr-1", "decigr-1", "deci-gr-1", "dgr⁻¹", "decigr⁻¹", "deci-gr⁻¹", "dg⁻¹", "decig⁻¹", "deci-g⁻¹"], "followed_by": None},
    "per_cg": {"dim": "mass", "degree": -1, "scale": 1e2, "terms": ["cg-1", "centig-1", "centi-g-1", "cgr-1", "centigr-1", "centi-gr-1", "cgr⁻¹", "centigr⁻¹", "centi-gr⁻¹", "cg⁻¹", "centig⁻¹", "centi-g⁻¹"], "followed_by": None},
    "per_mg": {"dim": "mass", "degree": -1, "scale": 1e3, "terms": ["mg-1", "millig-1", "milli-g-1", "milig-1", "mili-g-1", "mgr-1", "milligr-1", "milli-gr-1", "miligr-1", "mili-gr-1", "mgr⁻¹", "milligr⁻¹", "milli-gr⁻¹", "miligr⁻¹", "mili-gr⁻¹", "mg⁻¹", "millig⁻¹", "milli-g⁻¹", "milig⁻¹", "mili-g⁻¹"], "followed_by": None},
    "per_μg": {"dim": "mass", "degree": -1, "scale": 1e6, "terms": ["μg-1", "ug-1", "µg-1", "microg-1", "micro-g-1", "μgr-1", "ugr-1", "µgr-1", "microgr-1", "micro-gr-1", "μgr⁻¹", "ugr⁻¹", "µgr⁻¹", "microgr⁻¹", "micro-gr⁻¹", "μg⁻¹", "ug⁻¹", "µg⁻¹", "microg⁻¹", "micro-g⁻¹"], "followed_by": None},
    "per_ng": {"dim": "mass", "degree": -1, "scale": 1e9, "terms": ["ng-1", "nanog-1", "nano-g-1", "ngr-1", "nanogr-1", "nano-gr-1", "ngr⁻¹", "nanogr⁻¹", "nano-gr⁻¹", "ng⁻¹", "nanog⁻¹", "nano-g⁻¹"], "followed_by": None},
    "per_pg": {"dim": "mass", "degree": -1, "scale": 1e12, "terms": ["pg-1", "picog-1", "pico-g-1", "pgr-1", "picogr-1", "pico-gr-1", "pgr⁻¹", "picogr⁻¹", "pico-gr⁻¹", "pg⁻¹", "picog⁻¹", "pico-g⁻¹"], "followed_by": None},
    "per_fg": {"dim": "mass", "degree": -1, "scale": 1e15, "terms": ["fg-1", "femtog-1", "femto-g-1", "fgr-1", "femtogr-1", "femto-gr-1", "fgr⁻¹", "femtogr⁻¹", "femto-gr⁻¹", "fg⁻¹", "femtog⁻¹", "femto-g⁻¹"], "followed_by": None},

    # Length
    "km":  {"dim": "length", "degree": 1, "scale": 1e5, "terms": ["kmetre", "kilometre", "kilo-metre", "kmetres", "kilometres", "kilo-metres", "km", "kilom", "kilo-m"], "followed_by": None},
    "hm":  {"dim": "length", "degree": 1, "scale": 1e4, "terms": ["hmetre", "hectometre", "hecto-metre", "hmetres", "hectometres", "hecto-metres", "hm", "hectom", "hecto-m"], "followed_by": None},
    "dam": {"dim": "length", "degree": 1, "scale": 1e3, "terms": ["dametre", "decametre", "deca-metre", "dametres", "decametres", "deca-metres", "dam", "decam", "deca-m"], "followed_by": None},
    "m":   {"dim": "length", "degree": 1, "scale": 1e2, "terms": ["metre", "metres", "m"], "followed_by": "cm"},
    "dm":  {"dim": "length", "degree": 1, "scale": 1e1, "terms": ["dmetre", "decimetre", "deci-metre", "dmetres", "decimetres", "deci-metres", "dm", "decim", "deci-m"], "followed_by": None},
    "cm":  {"dim": "length", "degree": 1, "scale": 1e0, "terms": ["cmetre", "centimetre", "centi-metre", "cmetres", "centimetres", "centi-metres", "cm", "centim", "centi-m"], "followed_by": None},
    "mm":  {"dim": "length", "degree": 1, "scale": 1e-1, "terms": ["mmetre", "millimetre", "milli-metre", "milimetre", "mili-metre", "mmetres", "millimetres", "milli-metres", "milimetres", "mili-metres", "mm", "millim", "milli-m", "milim", "mili-m"], "followed_by": None},
    "μm":  {"dim": "length", "degree": 1, "scale": 1e-4, "terms": ["μmetre", "umetre", "µmetre", "micrometre", "micro-metre", "μmetres", "umetres", "µmetres", "micrometres", "micro-metres", "μm", "um", "µm", "microm", "micro-m"], "followed_by": None},
    "nm":  {"dim": "length", "degree": 1, "scale": 1e-7, "terms": ["nmetre", "nanometre", "nano-metre", "nmetres", "nanometres", "nano-metres", "nm", "nanom", "nano-m"], "followed_by": None},
    "pm":  {"dim": "length", "degree": 1, "scale": 1e-10, "terms": ["pmetre", "picometre", "pico-metre", "pmetres", "picometres", "pico-metres", "pm", "picom", "pico-m"], "followed_by": None},
    "fm":  {"dim": "length", "degree": 1, "scale": 1e-13, "terms": ["fmetre", "femtometre", "femto-metre", "fmetres", "femtometres", "femto-metres", "fm", "femtom", "femto-m"], "followed_by": None},

    # Length^-1
    "per_km":  {"dim": "length", "degree": -1, "scale": 1e-5, "terms": ["km-1", "kilom-1", "kilo-m-1", "km⁻¹", "kilom⁻¹", "kilo-m⁻¹"], "followed_by": None},
    "per_hm":  {"dim": "length", "degree": -1, "scale": 1e-4, "terms": ["hm-1", "hectom-1", "hecto-m-1", "hm⁻¹", "hectom⁻¹", "hecto-m⁻¹"], "followed_by": None},
    "per_dam": {"dim": "length", "degree": -1, "scale": 1e-3, "terms": ["dam-1", "decam-1", "deca-m-1", "dam⁻¹", "decam⁻¹", "deca-m⁻¹"], "followed_by": None},
    "per_m":   {"dim": "length", "degree": -1, "scale": 1e-2, "terms": ["m-1", "m⁻¹"], "followed_by": None},
    "per_dm":  {"dim": "length", "degree": -1, "scale": 1e-1, "terms": ["dm-1", "decim-1", "deci-m-1", "dm⁻¹", "decim⁻¹", "deci-m⁻¹"], "followed_by": None},
    "per_cm":  {"dim": "length", "degree": -1, "scale": 1e0, "terms": ["cm-1", "centim-1", "centi-m-1", "cm⁻¹", "centim⁻¹", "centi-m⁻¹"], "followed_by": None},
    "per_mm":  {"dim": "length", "degree": -1, "scale": 1e1, "terms": ["mm-1", "millim-1", "milli-m-1", "milim-1", "mili-m-1", "mm⁻¹", "millim⁻¹", "milli-m⁻¹", "milim⁻¹", "mili-m⁻¹"], "followed_by": None},
    "per_μm":  {"dim": "length", "degree": -1, "scale": 1e4, "terms": ["μm-1", "um-1", "µm-1", "microm-1", "micro-m-1", "μm⁻¹", "um⁻¹", "µm⁻¹", "microm⁻¹", "micro-m⁻¹"], "followed_by": None},
    "per_nm":  {"dim": "length", "degree": -1, "scale": 1e7, "terms": ["nm-1", "nanom-1", "nano-m-1", "nm⁻¹", "nanom⁻¹", "nano-m⁻¹"], "followed_by": None},
    "per_pm":  {"dim": "length", "degree": -1, "scale": 1e10, "terms": ["pm-1", "picom-1", "pico-m-1", "pm⁻¹", "picom⁻¹", "pico-m⁻¹"], "followed_by": None},
    "per_fm":  {"dim": "length", "degree": -1, "scale": 1e13, "terms": ["fm-1", "femtom-1", "femto-m-1", "fm⁻¹", "femtom⁻¹", "femto-m⁻¹"], "followed_by": None},

    # Length^2
    "km2":  {"dim": "length", "degree": 2, "scale": 1e10, "terms": ["km2", "kilom2", "kilo-m2", "km²", "kilom²", "kilo-m²"], "followed_by": None},
    "hm2":  {"dim": "length", "degree": 2, "scale": 1e8, "terms": ["hm2", "hectom2", "hecto-m2", "hm²", "hectom²", "hecto-m²"], "followed_by": None},
    "dam2": {"dim": "length", "degree": 2, "scale": 1e6, "terms": ["dam2", "decam2", "deca-m2", "dam²", "decam²", "deca-m²"], "followed_by": None},
    "m2":   {"dim": "length", "degree": 2, "scale": 1e4, "terms": ["m2", "m²"], "followed_by": None},
    "dm2":  {"dim": "length", "degree": 2, "scale": 1e2, "terms": ["dm2", "decim2", "deci-m2", "dm²", "decim²", "deci-m²"], "followed_by": None},
    "cm2":  {"dim": "length", "degree": 2, "scale": 1e0, "terms": ["cm2", "centim2", "centi-m2", "cm²", "centim²", "centi-m²"], "followed_by": None},
    "mm2":  {"dim": "length", "degree": 2, "scale": 1e-2, "terms": ["mm2", "millim2", "milli-m2", "milim2", "mili-m2", "mm²", "millim²", "milli-m²", "milim²", "mili-m²"], "followed_by": None},
    "μm2":  {"dim": "length", "degree": 2, "scale": 1e-8, "terms": ["μm2", "um2", "µm2", "microm2", "micro-m2", "μm²", "um²", "µm²", "microm²", "micro-m²"], "followed_by": None},
    "nm2":  {"dim": "length", "degree": 2, "scale": 1e-14, "terms": ["nm2", "nanom2", "nano-m2", "nm²", "nanom²", "nano-m²"], "followed_by": None},
    "pm2":  {"dim": "length", "degree": 2, "scale": 1e-20, "terms": ["pm2", "picom2", "pico-m2", "pm²", "picom²", "pico-m²"], "followed_by": None},
    "fm2":  {"dim": "length", "degree": 2, "scale": 1e-26, "terms": ["fm2", "femtom2", "femto-m2", "fm²", "femtom²", "femto-m²"], "followed_by": None},

    # Length^-2
    "per_km2":  {"dim": "length", "degree": -2, "scale": 1e-10, "terms": ["km-2", "kilom-2", "kilo-m-2", "km⁻²", "kilom⁻²", "kilo-m⁻²"], "followed_by": None},
    "per_hm2":  {"dim": "length", "degree": -2, "scale": 1e-8, "terms": ["hm-2", "hectom-2", "hecto-m-2", "hm⁻²", "hectom⁻²", "hecto-m⁻²"], "followed_by": None},
    "per_dam2": {"dim": "length", "degree": -2, "scale": 1e-6, "terms": ["dam-2", "decam-2", "deca-m-2", "dam⁻²", "decam⁻²", "deca-m⁻²"], "followed_by": None},
    "per_m2":   {"dim": "length", "degree": -2, "scale": 1e-4, "terms": ["m-2", "m⁻²"], "followed_by": None},
    "per_dm2":  {"dim": "length", "degree": -2, "scale": 1e-2, "terms": ["dm-2", "decim-2", "deci-m-2", "dm⁻²", "decim⁻²", "deci-m⁻²"], "followed_by": None},
    "per_cm2":  {"dim": "length", "degree": -2, "scale": 1e0, "terms": ["cm-2", "centim-2", "centi-m-2", "cm⁻²", "centim⁻²", "centi-m⁻²"], "followed_by": None},
    "per_mm2":  {"dim": "length", "degree": -2, "scale": 1e2, "terms": ["mm-2", "millim-2", "milli-m-2", "milim-2", "mili-m-2", "mm⁻²", "millim⁻²", "milli-m⁻²", "milim⁻²", "mili-m⁻²"], "followed_by": None},
    "per_μm2":  {"dim": "length", "degree": -2, "scale": 1e8, "terms": ["μm-2", "um-2", "µm-2", "microm-2", "micro-m-2", "μm⁻²", "um⁻²", "µm⁻²", "microm⁻²", "micro-m⁻²"], "followed_by": None},
    "per_nm2":  {"dim": "length", "degree": -2, "scale": 1e14, "terms": ["nm-2", "nanom-2", "nano-m-2", "nm⁻²", "nanom⁻²", "nano-m⁻²"], "followed_by": None},
    "per_pm2":  {"dim": "length", "degree": -2, "scale": 1e20, "terms": ["pm-2", "picom-2", "pico-m-2", "pm⁻²", "picom⁻²", "pico-m⁻²"], "followed_by": None},
    "per_fm2":  {"dim": "length", "degree": -2, "scale": 1e26, "terms": ["fm-2", "femtom-2", "femto-m-2", "fm⁻²", "femtom⁻²", "femto-m⁻²"], "followed_by": None},

    # Length^3
    "km3":    {"dim": "length", "degree": 3, "scale": 1e15, "terms": ["km3", "kilom3", "kilo-m3", "km³", "kilom³", "kilo-m³"], "followed_by": None},
    "hm3":    {"dim": "length", "degree": 3, "scale": 1e12, "terms": ["hm3", "hectom3", "hecto-m3", "hm³", "hectom³", "hecto-m³"], "followed_by": None},
    "dam3":   {"dim": "length", "degree": 3, "scale": 1e9, "terms": ["dam3", "decam3", "deca-m3", "dam³", "decam³", "deca-m³"], "followed_by": None},
    "m3":     {"dim": "length", "degree": 3, "scale": 1e6, "terms": ["m3", "m³"], "followed_by": None},
    "dm3":    {"dim": "length", "degree": 3, "scale": 1e3, "terms": ["dm3", "decim3", "deci-m3", "dm³", "decim³", "deci-m³"], "followed_by": None},
    "cm3":    {"dim": "length", "degree": 3, "scale": 1e0, "terms": ["cm3", "centim3", "centi-m3", "cm³", "centim³", "centi-m³"], "followed_by": None},
    "mm3":    {"dim": "length", "degree": 3, "scale": 1e-3, "terms": ["mm3", "millim3", "milli-m3", "milim3", "mili-m3", "mm³", "millim³", "milli-m³", "milim³", "mili-m³"], "followed_by": None},
    "μm3":    {"dim": "length", "degree": 3, "scale": 1e-12, "terms": ["μm3", "um3", "µm3", "microm3", "micro-m3", "μm³", "um³", "µm³", "microm³", "micro-m³"], "followed_by": None},
    "nm3":    {"dim": "length", "degree": 3, "scale": 1e-21, "terms": ["nm3", "nanom3", "nano-m3", "nm³", "nanom³", "nano-m³"], "followed_by": None},
    "pm3":    {"dim": "length", "degree": 3, "scale": 1e-30, "terms": ["pm3", "picom3", "pico-m3", "pm³", "picom³", "pico-m³"], "followed_by": None},
    "fm3":    {"dim": "length", "degree": 3, "scale": 1e-39, "terms": ["fm3", "femtom3", "femto-m3", "fm³", "femtom³", "femto-m³"], "followed_by": None},
    "cac":    {"dim": "length", "degree": 3, "scale": 5.0, "terms": ["cac", "c.a.c", "cuillere à café", "cuillères à café"], "followed_by": None},
    "goutte": {"dim": "length", "degree": 3, "scale": 0.05, "terms": ["gt", "goutte", "gouttes"], "followed_by": None},
    "µl":     {"dim": "length", "degree": 3, "scale": 1e-3, "terms": ["µl", "μl", "µlitre", "μlitres", "micro litre", "micro litres", "microlitre", "microlitres", "micro-litre", "micro-litres"], "followed_by": None},
    "ml":     {"dim": "length", "degree": 3, "scale": 1e0, "terms": ["mililitre", "millilitre", "mililitres", "millilitres", "ml"], "followed_by": None},
    "cl":     {"dim": "length", "degree": 3, "scale": 1e1, "terms": ["centilitre", "centilitres", "cl"], "followed_by": None},
    "dl":     {"dim": "length", "degree": 3, "scale": 1e2, "terms": ["decilitre", "decilitres", "dl"], "followed_by": None},
    "l":      {"dim": "length", "degree": 3, "scale": 1e3, "terms": ["litre", "litres", "l", "dm3"], "followed_by": "ml"},

    # Length^-3
    "per_km3":  {"dim": "length", "degree": -3, "scale": 1e-15, "terms": ["km-3", "kilom-3", "kilo-m-3", "km⁻³", "kilom⁻³", "kilo-m⁻³"], "followed_by": None},
    "per_hm3":  {"dim": "length", "degree": -3, "scale": 1e-12, "terms": ["hm-3", "hectom-3", "hecto-m-3", "hm⁻³", "hectom⁻³", "hecto-m⁻³"], "followed_by": None},
    "per_dam3": {"dim": "length", "degree": -3, "scale": 1e-9, "terms": ["dam-3", "decam-3", "deca-m-3", "dam⁻³", "decam⁻³", "deca-m⁻³"], "followed_by": None},
    "per_dm3":  {"dim": "length", "degree": -3, "scale": 1e-6, "terms": ["dm-3", "decim-3", "deci-m-3", "dm⁻³", "decim⁻³", "deci-m⁻³"], "followed_by": None},
    "per_m3":   {"dim": "length", "degree": -3, "scale": 1e-3, "terms": ["m-3", "m⁻³"], "followed_by": None},
    "per_cm3":  {"dim": "length", "degree": -3, "scale": 1e0, "terms": ["cm-3", "centim-3", "centi-m-3", "cm⁻³", "centim⁻³", "centi-m⁻³"], "followed_by": None},
    "per_mm3":  {"dim": "length", "degree": -3, "scale": 1e3, "terms": ["mm-3", "millim-3", "milli-m-3", "milim-3", "mili-m-3", "mm⁻³", "millim⁻³", "milli-m⁻³", "milim⁻³", "mili-m⁻³"], "followed_by": None},
    "per_μm3":  {"dim": "length", "degree": -3, "scale": 1e12, "terms": ["μm-3", "um-3", "µm-3", "microm-3", "micro-m-3", "μm⁻³", "um⁻³", "µm⁻³", "microm⁻³", "micro-m⁻³"], "followed_by": None},
    "per_nm3":  {"dim": "length", "degree": -3, "scale": 1e21, "terms": ["nm-3", "nanom-3", "nano-m-3", "nm⁻³", "nanom⁻³", "nano-m⁻³"], "followed_by": None},
    "per_pm3":  {"dim": "length", "degree": -3, "scale": 1e30, "terms": ["pm-3", "picom-3", "pico-m-3", "pm⁻³", "picom⁻³", "pico-m⁻³"], "followed_by": None},
    "per_fm3":  {"dim": "length", "degree": -3, "scale": 1e39, "terms": ["fm-3", "femtom-3", "femto-m-3", "fm⁻³", "femtom⁻³", "femto-m⁻³"], "followed_by": None},

    # Mole
    "kmol":  {"dim": "mole", "degree": 1, "scale": 1e3, "terms": ["kmol", "kilomol", "kilo-mol", "kmole", "kilomole", "kilo-mole", "kmoles", "kilomoles", "kilo-moles"], "followed_by": None},
    "hmol":  {"dim": "mole", "degree": 1, "scale": 1e2, "terms": ["hmol", "hectomol", "hecto-mol", "hmole", "hectomole", "hecto-mole", "hmoles", "hectomoles", "hecto-moles"], "followed_by": None},
    "damol": {"dim": "mole", "degree": 1, "scale": 1e1, "terms": ["damol", "decamol", "deca-mol", "damole", "decamole", "deca-mole", "damoles", "decamoles", "deca-moles"], "followed_by": None},
    "mol":   {"dim": "mole", "degree": 1, "scale": 1e0,  "terms": ["mol", "mole", "moles"], "followed_by": None},
    "dmol":  {"dim": "mole", "degree": 1, "scale": 1e-1, "terms": ["dmol", "decimol", "deci-mol", "dmole", "decimole", "deci-mole", "dmoles", "decimoles", "deci-moles"], "followed_by": None},
    "cmol":  {"dim": "mole", "degree": 1, "scale": 1e-2, "terms": ["cmol", "centimol", "centi-mol", "cmole", "centimole", "centi-mole", "cmoles", "centimoles", "centi-moles"], "followed_by": None},
    "mmol":  {"dim": "mole", "degree": 1, "scale": 1e-3, "terms": ["mmol", "millimol", "milli-mol", "milimol", "mili-mol", "mmole", "millimole", "milli-mole", "milimole", "mili-mole", "mmoles", "millimoles", "milli-moles", "milimoles", "mili-moles"], "followed_by": None},
    "μmol":  {"dim": "mole", "degree": 1, "scale": 1e-6, "terms": ["μmol", "umol", "µmol", "micromol", "micro-mol", "μmole", "umole", "µmole", "micromole", "micro-mole", "μmoles", "umoles", "µmoles", "micromoles", "micro-moles"], "followed_by": None},
    "nmol":  {"dim": "mole", "degree": 1, "scale": 1e-9, "terms": ["nmol", "nanomol", "nano-mol", "nmole", "nanomole", "nano-mole", "nmoles", "nanomoles", "nano-moles"], "followed_by": None},
    "pmol":  {"dim": "mole", "degree": 1, "scale": 1e-12, "terms": ["pmol", "picomol", "pico-mol", "pmole", "picomole", "pico-mole", "pmoles", "picomoles", "pico-moles"], "followed_by": None},
    "fmol":  {"dim": "mole", "degree": 1, "scale": 1e-15, "terms": ["fmol", "femtomol", "femto-mol", "fmole", "femtomole", "femto-mole", "fmoles", "femtomoles", "femto-moles"], "followed_by": None},

    # UI
    "kui":  {"dim": "ui", "degree": 1, "scale": 1e3, "terms": ["kui", "kiloui", "kilo-ui", "ku", "kilou", "kilo-u"], "followed_by": None},
    "hui":  {"dim": "ui", "degree": 1, "scale": 1e2, "terms": ["hui", "hectoui", "hecto-ui", "hu", "hectou", "hecto-u"], "followed_by": None},
    "daui": {"dim": "ui", "degree": 1, "scale": 1e1, "terms": ["daui", "decaui", "deca-ui", "dau", "decau", "deca-u"], "followed_by": None},
    "ui":   {"dim": "ui", "degree": 1, "scale": 1e0, "terms": ["ui", "u"], "followed_by": None},
    "dui":  {"dim": "ui", "degree": 1, "scale": 1e-1, "terms": ["dui", "deciui", "deci-ui", "du", "deciu", "deci-u"], "followed_by": None},
    "cui":  {"dim": "ui", "degree": 1, "scale": 1e-2, "terms": ["cui", "centiui", "centi-ui", "cu", "centiu", "centi-u"], "followed_by": None},
    "mui":  {"dim": "ui", "degree": 1, "scale": 1e-3, "terms": ["mui", "milliui", "milli-ui", "miliui", "mili-ui", "mu", "milliu", "milli-u", "miliu", "mili-u"], "followed_by": None},
    "μui":  {"dim": "ui", "degree": 1, "scale": 1e-6, "terms": ["μui", "uui", "µui", "microui", "micro-ui", "μu", "uu", "µu", "microu", "micro-u"], "followed_by": None},
    "nui":  {"dim": "ui", "degree": 1, "scale": 1e-9, "terms": ["nui", "nanoui", "nano-ui", "nu", "nanou", "nano-u"], "followed_by": None},
    "pui":  {"dim": "ui", "degree": 1, "scale": 1e-12, "terms": ["pui", "picoui", "pico-ui", "pu", "picou", "pico-u"], "followed_by": None},
    "fui":  {"dim": "ui", "degree": 1, "scale": 1e-15, "terms": ["fui", "femtoui", "femto-ui", "fu", "femtou", "femto-u"], "followed_by": None},

    # UI^-1
    "per_kui":  {"dim": "ui", "degree": -1, "scale": 1e-3, "terms": ["kui-1", "kiloui-1", "kilo-ui-1", "kui⁻¹", "kiloui⁻¹", "kilo-ui⁻¹"], "followed_by": None},
    "per_hui":  {"dim": "ui", "degree": -1, "scale": 1e-2, "terms": ["hui-1", "hectoui-1", "hecto-ui-1", "hui⁻¹", "hectoui⁻¹", "hecto-ui⁻¹"], "followed_by": None},
    "per_daui": {"dim": "ui", "degree": -1, "scale": 1e-1, "terms": ["daui-1", "decaui-1", "deca-ui-1", "daui⁻¹", "decaui⁻¹", "deca-ui⁻¹"], "followed_by": None},
    "per_ui":   {"dim": "ui", "degree": -1, "scale": 1e0, "terms": ["ui-1", "ui⁻¹"], "followed_by": None},
    "per_dui":  {"dim": "ui", "degree": -1, "scale": 1e1, "terms": ["dui-1", "deciui-1", "deci-ui-1", "dui⁻¹", "deciui⁻¹", "deci-ui⁻¹"], "followed_by": None},
    "per_cui":  {"dim": "ui", "degree": -1, "scale": 1e2, "terms": ["cui-1", "centiui-1", "centi-ui-1", "cui⁻¹", "centiui⁻¹", "centi-ui⁻¹"], "followed_by": None},
    "per_mui":  {"dim": "ui", "degree": -1, "scale": 1e3, "terms": ["mui-1", "milliui-1", "milli-ui-1", "miliui-1", "mili-ui-1", "mui⁻¹", "milliui⁻¹", "milli-ui⁻¹", "miliui⁻¹", "mili-ui⁻¹"], "followed_by": None},
    "per_μui":  {"dim": "ui", "degree": -1, "scale": 1e6, "terms": ["μui-1", "uui-1", "µui-1", "microui-1", "micro-ui-1", "μui⁻¹", "uui⁻¹", "µui⁻¹", "microui⁻¹", "micro-ui⁻¹"], "followed_by": None},
    "per_nui":  {"dim": "ui", "degree": -1, "scale": 1e9, "terms": ["nui-1", "nanoui-1", "nano-ui-1", "nui⁻¹", "nanoui⁻¹", "nano-ui⁻¹"], "followed_by": None},
    "per_pui":  {"dim": "ui", "degree": -1, "scale": 1e12, "terms": ["pui-1", "picoui-1", "pico-ui-1", "pui⁻¹", "picoui⁻¹", "pico-ui⁻¹"], "followed_by": None},
    "per_fui":  {"dim": "ui", "degree": -1, "scale": 1e15, "terms": ["fui-1", "femtoui-1", "femto-ui-1", "fui⁻¹", "femtoui⁻¹", "femto-ui⁻¹"], "followed_by": None},

    # Pressure
    "kPa":  {"dim": "pressure", "degree": 1, "scale": 1e3, "terms": ["kPa", "kiloPa", "kilo-Pa"], "followed_by": None},
    "hPa":  {"dim": "pressure", "degree": 1, "scale": 1e2, "terms": ["hPa", "hectoPa", "hecto-Pa"], "followed_by": None},
    "daPa": {"dim": "pressure", "degree": 1, "scale": 1e1, "terms": ["daPa", "decaPa", "deca-Pa"], "followed_by": None},
    "Pa":   {"dim": "pressure", "degree": 1, "scale": 1e0, "terms": ["Pa"], "followed_by": None},
    "dPa":  {"dim": "pressure", "degree": 1, "scale": 1e-1, "terms": ["dPa", "deciPa", "deci-Pa"], "followed_by": None},
    "cPa":  {"dim": "pressure", "degree": 1, "scale": 1e-2, "terms": ["cPa", "centiPa", "centi-Pa"], "followed_by": None},
    "mPa":  {"dim": "pressure", "degree": 1, "scale": 1e-3, "terms": ["mPa", "milliPa", "milli-Pa", "miliPa", "mili-Pa"], "followed_by": None},
    "μPa":  {"dim": "pressure", "degree": 1, "scale": 1e-6, "terms": ["μPa", "uPa", "µPa", "microPa", "micro-Pa"], "followed_by": None},
    "nPa":  {"dim": "pressure", "degree": 1, "scale": 1e-9, "terms": ["nPa", "nanoPa", "nano-Pa"], "followed_by": None},
    "pPa":  {"dim": "pressure", "degree": 1, "scale": 1e-12, "terms": ["pPa", "picoPa", "pico-Pa"], "followed_by": None},
    "fPa":  {"dim": "pressure", "degree": 1, "scale": 1e-15, "terms": ["fPa", "femtoPa", "femto-Pa"], "followed_by": None},
    "mmHg": {"dim": "pressure", "degree": 1, "scale": 133, "terms": ["mmHg"], "followed_by": None},

    # Percent (special unit)
    "%": {"dim": "%", "degree": 1, "scale": 1, "terms": ["%"], "followed_by": None},

    # Logarithm (special unit)
    "log": {"dim": "log", "degree": 1, "scale": 1, "terms": ["log"], "followed_by": None},

    # Time
    "year":       {"dim": "time", "degree": 1, "scale": 31557600.0, "terms": ["an", "année", "ans", "années"], "followed_by": None},
    "month":      {"dim": "time", "degree": 1, "scale": 2628002.88, "terms": ["mois"], "followed_by": None},
    "week":       {"dim": "time", "degree": 1, "scale": 604800, "terms": ["semaine", "semaines", "sem"], "followed_by": None},
    "day":        {"dim": "time", "degree": 1, "scale": 86400, "terms": ["jour", "jours", "j"], "followed_by": None},
    "hour":       {"dim": "time", "degree": 1, "scale": 3600, "terms": ["heure", "heures", "h"], "followed_by": "minute"},
    "minute":     {"dim": "time", "degree": 1, "scale": 60, "terms": ["mn", "min", "minute", "minutes"], "followed_by": "second"},
    "second":     {"dim": "time", "degree": 1, "scale": 1, "terms": ["seconde", "secondes", "s"], "followed_by": None},
    "arc-minute": {"dim": "time", "degree": 1, "scale": 60, "terms": ["'"], "followed_by": "arc-second"},
    "arc-second": {"dim": "time", "degree": 1, "scale": 1, "terms": ['"', "''"], "followed_by": None},
    "degree":     {"dim": "time", "degree": 1, "scale": 3600, "terms": ["degre", "°", "deg"], "followed_by": "arc-minute"},

    # Temperature
    "celsius": {"dim": "temperature", "degree": 1, "scale": 1, "terms": ["°C", "° celsius", "celsius"], "followed_by": None},

    # Count
    "x10*1":  {"dim": "count", "degree": 0, "scale": 1e1, "terms": ["x10*1"], "followed_by": None},
    "x10*2":  {"dim": "count", "degree": 0, "scale": 1e2, "terms": ["x10*2"], "followed_by": None},
    "x10*3":  {"dim": "count", "degree": 0, "scale": 1e3, "terms": ["x10*3"], "followed_by": None},
    "x10*4":  {"dim": "count", "degree": 0, "scale": 1e4, "terms": ["x10*4"], "followed_by": None},
    "x10*5":  {"dim": "count", "degree": 0, "scale": 1e5, "terms": ["x10*5"], "followed_by": None},
    "x10*6":  {"dim": "count", "degree": 0, "scale": 1e6, "terms": ["x10*6"], "followed_by": None},
    "x10*7":  {"dim": "count", "degree": 0, "scale": 1e7, "terms": ["x10*7"], "followed_by": None},
    "x10*8":  {"dim": "count", "degree": 0, "scale": 1e8, "terms": ["x10*8"], "followed_by": None},
    "x10*9":  {"dim": "count", "degree": 0, "scale": 1e9, "terms": ["x10*9"], "followed_by": None},
    "x10*10": {"dim": "count", "degree": 0, "scale": 1e10, "terms": ["x10*10"], "followed_by": None},
    "x10*11": {"dim": "count", "degree": 0, "scale": 1e11, "terms": ["x10*11"], "followed_by": None},
    "x10*12": {"dim": "count", "degree": 0, "scale": 1e12, "terms": ["x10*12"], "followed_by": None},
    "x10*13": {"dim": "count", "degree": 0, "scale": 1e13, "terms": ["x10*13"], "followed_by": None},
    "x10*14": {"dim": "count", "degree": 0, "scale": 1e14, "terms": ["x10*14"], "followed_by": None},
    "x10*15": {"dim": "count", "degree": 0, "scale": 1e15, "terms": ["x10*15"], "followed_by": None},
}
# fmt: on
